from django.test import TestCase

from ... import EthereumClient
from ...oracles import SushiswapOracle
from ...oracles.utils import get_decimals as oracles_get_decimals
from ..ethereum_test_case import EthereumTestCaseMixin
from ..test_oracles import (
    dai_token_mainnet_address,
    usdt_token_mainnet_address,
    wbtc_token_mainnet_address,
    weth_token_mainnet_address,
)
from ..utils import just_test_if_mainnet_node


class TestSushiSwapOracle(EthereumTestCaseMixin, TestCase):
    def test_get_price(self):
        oracles_get_decimals.cache_clear()
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(SushiswapOracle.is_available(ethereum_client))
        sushiswap_oracle = SushiswapOracle(ethereum_client)

        price = sushiswap_oracle.get_price(
            wbtc_token_mainnet_address, weth_token_mainnet_address
        )
        self.assertGreater(price, 0)
        self.assertEqual(oracles_get_decimals.cache_info().currsize, 2)

        # Test with 2 stablecoins
        price = sushiswap_oracle.get_price(
            dai_token_mainnet_address, usdt_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)
        self.assertEqual(oracles_get_decimals.cache_info().currsize, 4)

        self.assertEqual(oracles_get_decimals.cache_info().hits, 0)
        self.assertEqual(
            oracles_get_decimals(dai_token_mainnet_address, ethereum_client), 18
        )
        self.assertEqual(
            oracles_get_decimals(usdt_token_mainnet_address, ethereum_client), 6
        )
        self.assertEqual(oracles_get_decimals.cache_info().hits, 2)

        price = sushiswap_oracle.get_price(
            usdt_token_mainnet_address, dai_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)
        self.assertEqual(oracles_get_decimals.cache_info().currsize, 4)
        self.assertEqual(oracles_get_decimals.cache_info().hits, 4)
        oracles_get_decimals.cache_clear()



from unittest import TestCase

from gnosis.eth.eip712 import eip712_encode_hash


class TestEIP712(TestCase):
    address = "0x8e12f01DAE5FE7f1122Dc42f2cB084F2f9E8aA03"
    types = {
        "EIP712Domain": [
            {"name": "name", "type": "string"},
            {"name": "version", "type": "string"},
            {"name": "chainId", "type": "uint256"},
            {"name": "verifyingContract", "type": "address"},
        ],
        "Mailbox": [
            {"name": "owner", "type": "address"},
            {"name": "messages", "type": "Message[]"},
        ],
        "Message": [
            {"name": "sender", "type": "address"},
            {"name": "subject", "type": "string"},
            {"name": "isSpam", "type": "bool"},
            {"name": "body", "type": "string"},
        ],
    }

    msgs = [
        {
            "sender": address,
            "subject": "Hello World",
            "body": "The sparrow flies at midnight.",
            "isSpam": False,
        },
        {
            "sender": address,
            "subject": "You may have already Won! :dumb-emoji:",
            "body": "Click here for sweepstakes!",
            "isSpam": True,
        },
    ]

    mailbox = {"owner": address, "messages": msgs}

    def test_eip712_encode_hash(self):
        for value in [
            {},
            None,
        ]:
            with self.assertRaises(ValueError):
                eip712_encode_hash(value)

        wrong_types = {
            "EIP712Domain": [
                {"name": "name", "type": "stringa"},
                {"name": "version", "type": "bstring"},
                {"name": "chainId", "type": "aaauint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            "Mailbox": [
                {"name": "owner", "type": "address"},
                {"name": "messages", "type": "Message[]"},
            ],
            "Message": [
                {"name": "sender", "type": "address"},
                {"name": "subject", "type": "string"},
                {"name": "isSpam", "type": "bool"},
                {"name": "body", "type": "string"},
            ],
        }

        payload = {
            "types": wrong_types,
            "primaryType": "Mailbox",
            "domain": {
                "name": "MyDApp",
                "version": "3.0",
                "chainId": 41,
                "verifyingContract": self.address,
            },
            "message": self.mailbox,
        }

        with self.assertRaises(ValueError):
            eip712_encode_hash(payload)

        payload["types"] = self.types
        self.assertEqual(
            eip712_encode_hash(payload).hex(),
            "0x7c02fe79823722257b42ea95720e7dd31d51c3f6769dc0f56a271800dd030ef1",
        )

    def test_eip712_encode_hash_string_uint(self):
        # test string uint (chainId)
        payload = {
            "types": self.types,
            "primaryType": "Mailbox",
            "domain": {
                "name": "MyDApp",
                "version": "3.0",
                "chainId": "41",
                "verifyingContract": self.address,
            },
            "message": self.mailbox,
        }
        self.assertEqual(
            eip712_encode_hash(payload).hex(),
            "0x7c02fe79823722257b42ea95720e7dd31d51c3f6769dc0f56a271800dd030ef1",
        )

    def test_eip712_encode_hash_string_bytes(self):
        types_with_bytes = {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
            ],
            "Message": [
                {"name": "oneByte", "type": "bytes1"},
                {"name": "maxByte", "type": "bytes32"},
            ],
        }
        payload = {
            "types": types_with_bytes,
            "primaryType": "Message",
            "domain": {
                "name": "MyDApp",
            },
            "message": {
                "oneByte": "0x01",
                "maxByte": "0x6214da6089b8d8aaa6e6268977746aa0af19fd1ef5d56e225bb3390a697c3ec1",
            },
        }

        self.assertEqual(
            eip712_encode_hash(payload).hex(),
            "0x2950cf06416c6c20059f24a965e3baf51a24f4ef49a1e7b1a47ee13ee08cde1f",
        )

    def test_eip712_encode_nested_with_array(self):
        payload = {
            "types": {
                "Nested": [
                    {"name": "nestedString", "type": "string"},
                    {"name": "nestedAddress", "type": "address"},
                    {"name": "nestedUint256", "type": "uint256"},
                    {"name": "nestedUint32", "type": "uint32"},
                    {"name": "nestedBytes32", "type": "bytes32"},
                    {"name": "nestedBoolean", "type": "bool"},
                ],
                "Example": [
                    {"name": "testString", "type": "string"},
                    {"name": "testAddress", "type": "address"},
                    {"name": "testUint256", "type": "uint256"},
                    {"name": "testUint32", "type": "uint32"},
                    {"name": "testBytes32", "type": "bytes32"},
                    {"name": "testBoolean", "type": "bool"},
                    {"name": "testNested", "type": "Nested"},
                    {"name": "testNestedArray", "type": "Nested[]"},
                ],
                "EIP712Domain": [
                    {"name": "name", "type": "string"},
                    {"name": "version", "type": "string"},
                    {"name": "chainId", "type": "uint256"},
                    {"name": "verifyingContract", "type": "address"},
                ],
            },
            "domain": {
                "name": "EIP-1271 Example DApp",
                "version": "1.0",
                "chainId": "5",
                "verifyingContract": "0xaecdfd3a19f777f0c03e6bf99aafb59937d6467b",
            },
            "primaryType": "Example",
            "message": {
                "testString": "Hello Deeeeeemo",
                "testAddress": "0xaecdfd3a19f777f0c03e6bf99aafb59937d6467b",
                "testUint256": "115792089237316195423570985008687907853269984665640564039457584007908834671663",
                "testUint32": "123",
                "testBytes32": "0x00000000000000000000000000000000000000000000000000000000deadbeef",
                "testBoolean": True,
                "testNested": {
                    "nestedString": "Hello Deeeeeemo",
                    "nestedAddress": "0x0000000000000000000000000000000000000002",
                    "nestedUint256": "0",
                    "nestedUint32": "1",
                    "nestedBytes32": "0x000000000000000000000000000000000000000000000000000000000000da7a",
                    "nestedBoolean": False,
                },
                "testNestedArray": [
                    {
                        "nestedString": "Hello Deeeeeemo",
                        "nestedAddress": "0x0000000000000000000000000000000000000002",
                        "nestedUint256": "0",
                        "nestedUint32": "1",
                        "nestedBytes32": "0x000000000000000000000000000000000000000000000000000000000000da7a",
                        "nestedBoolean": False,
                    },
                    {
                        "nestedString": "Hello Deeeeeemo",
                        "nestedAddress": "0x0000000000000000000000000000000000000002",
                        "nestedUint256": "0",
                        "nestedUint32": "1",
                        "nestedBytes32": "0x000000000000000000000000000000000000000000000000000000000000da7a",
                        "nestedBoolean": False,
                    },
                ],
            },
        }
        self.assertEqual(
            eip712_encode_hash(payload).hex(),
            "0x2f6856dbd51836973c1e61852b64949556aa2e7f253d9e20e682f9a02d436791",
        )

    def test_eip712_encode_nested_with_empty_array(self):
        payload = {
            "types": {
                "Nested": [
                    {"name": "nestedString", "type": "string"},
                    {"name": "nestedAddress", "type": "address"},
                    {"name": "nestedUint256", "type": "uint256"},
                    {"name": "nestedUint32", "type": "uint32"},
                    {"name": "nestedBytes32", "type": "bytes32"},
                    {"name": "nestedBoolean", "type": "bool"},
                ],
                "Example": [
                    {"name": "testString", "type": "string"},
                    {"name": "testAddress", "type": "address"},
                    {"name": "testUint256", "type": "uint256"},
                    {"name": "testUint32", "type": "uint32"},
                    {"name": "testBytes32", "type": "bytes32"},
                    {"name": "testBoolean", "type": "bool"},
                    {"name": "testNested", "type": "Nested"},
                    {"name": "testNestedArray", "type": "Nested[]"},
                ],
                "EIP712Domain": [
                    {"name": "name", "type": "string"},
                    {"name": "version", "type": "string"},
                    {"name": "chainId", "type": "uint256"},
                    {"name": "verifyingContract", "type": "address"},
                ],
            },
            "domain": {
                "name": "EIP-1271 Example DApp",
                "version": "1.0",
                "chainId": "5",
                "verifyingContract": "0xaecdfd3a19f777f0c03e6bf99aafb59937d6467b",
            },
            "primaryType": "Example",
            "message": {
                "testString": "Hello Deeeeeemo",
                "testAddress": "0xaecdfd3a19f777f0c03e6bf99aafb59937d6467b",
                "testUint256": "115792089237316195423570985008687907853269984665640564039457584007908834671663",
                "testUint32": "123",
                "testBytes32": "0x00000000000000000000000000000000000000000000000000000000deadbeef",
                "testBoolean": True,
                "testNested": {
                    "nestedString": "Hello Deeeeeemo",
                    "nestedAddress": "0x0000000000000000000000000000000000000002",
                    "nestedUint256": "0",
                    "nestedUint32": "1",
                    "nestedBytes32": "0x000000000000000000000000000000000000000000000000000000000000da7a",
                    "nestedBoolean": False,
                },
                "testNestedArray": [],
            },
        }
        self.assertEqual(
            eip712_encode_hash(payload).hex(),
            "0x5dd3156111fb5e400606d4dd75ff097e36eb56614c84924b5eb6d1cf1b5038cf",
        )

    def test_eip712_encode_nested_without_array(self):
        payload = {
            "types": {
                "Nested": [
                    {"name": "nestedString", "type": "string"},
                    {"name": "nestedAddress", "type": "address"},
                    {"name": "nestedUint256", "type": "uint256"},
                    {"name": "nestedUint32", "type": "uint32"},
                    {"name": "nestedBytes32", "type": "bytes32"},
                    {"name": "nestedBoolean", "type": "bool"},
                ],
                "Example": [
                    {"name": "testString", "type": "string"},
                    {"name": "testAddress", "type": "address"},
                    {"name": "testUint256", "type": "uint256"},
                    {"name": "testUint32", "type": "uint32"},
                    {"name": "testBytes32", "type": "bytes32"},
                    {"name": "testBoolean", "type": "bool"},
                    {"name": "testNested", "type": "Nested"},
                ],
                "EIP712Domain": [
                    {"name": "name", "type": "string"},
                    {"name": "version", "type": "string"},
                    {"name": "chainId", "type": "uint256"},
                    {"name": "verifyingContract", "type": "address"},
                ],
            },
            "domain": {
                "name": "EIP-1271 Example DApp",
                "version": "1.0",
                "chainId": "5",
                "verifyingContract": "0xaecdfd3a19f777f0c03e6bf99aafb59937d6467b",
            },
            "primaryType": "Example",
            "message": {
                "testString": "Hello Deeeeeemo",
                "testAddress": "0xaecdfd3a19f777f0c03e6bf99aafb59937d6467b",
                "testUint256": "115792089237316195423570985008687907853269984665640564039457584007908834671663",
                "testUint32": "123",
                "testBytes32": "0x00000000000000000000000000000000000000000000000000000000deadbeef",
                "testBoolean": True,
                "testNested": {
                    "nestedString": "Hello Deeeeeemo",
                    "nestedAddress": "0x0000000000000000000000000000000000000002",
                    "nestedUint256": "0",
                    "nestedUint32": "1",
                    "nestedBytes32": "0x000000000000000000000000000000000000000000000000000000000000da7a",
                    "nestedBoolean": False,
                },
            },
        }

        self.assertEqual(
            eip712_encode_hash(payload).hex(),
            "0x9a55335a1d86221594e96018fc3df611a1485d95b5b2afbef1540ac51f63d249",
        )



from django.db import models

from ..models import (
    EthereumAddressField,
    EthereumAddressV2Field,
    Keccak256Field,
    Sha3HashField,
    Uint32Field,
    Uint96Field,
    Uint256Field,
)


class EthereumAddress(models.Model):
    value = EthereumAddressField(null=True)


class EthereumAddressV2(models.Model):
    value = EthereumAddressV2Field(null=True)


class Uint256(models.Model):
    value = Uint256Field(null=True)


class Uint96(models.Model):
    value = Uint96Field(null=True)


class Uint32(models.Model):
    value = Uint32Field(null=True)


class Sha3Hash(models.Model):
    value = Sha3HashField(null=True)


class Keccak256Hash(models.Model):
    value = Keccak256Field(null=True)



import logging
import os
from typing import Dict, List, Optional

from eth_account import Account
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3.contract import Contract
from web3.types import Wei

from gnosis.eth.constants import NULL_ADDRESS, SAFE_SINGLETON_FACTORY_DEPLOYER_ADDRESS
from gnosis.eth.contracts import (
    get_compatibility_fallback_handler_contract,
    get_multi_send_contract,
    get_proxy_factory_contract,
    get_safe_V1_0_0_contract,
    get_safe_V1_1_1_contract,
    get_safe_V1_3_0_contract,
    get_safe_V1_4_1_contract,
    get_sign_message_lib_contract,
    get_simulate_tx_accessor_V1_4_1_contract,
)
from gnosis.eth.exceptions import ContractAlreadyDeployed
from gnosis.eth.tests.ethereum_test_case import EthereumTestCaseMixin
from gnosis.eth.tests.utils import send_tx
from gnosis.eth.utils import get_empty_tx_params
from gnosis.safe import Safe
from gnosis.safe.multi_send import MultiSend
from gnosis.safe.proxy_factory import ProxyFactory, ProxyFactoryV141

from ..safe import SafeV001, SafeV100, SafeV111, SafeV130, SafeV141

logger = logging.getLogger(__name__)


class SafeTestCaseMixin(EthereumTestCaseMixin):
    compatibility_fallback_handler: Contract
    multi_send: MultiSend
    multi_send_contract: Contract
    proxy_factory: ProxyFactory
    proxy_factory_contract: Contract
    safe_contract_V1_4_1: Contract
    safe_contract_V0_0_1: Contract
    safe_contract_V1_0_0: Contract
    safe_contract_V1_1_1: Contract
    safe_contract_V1_3_0: Contract
    safe_contract_address: ChecksumAddress
    simulate_tx_accessor_V1_4_1: Contract

    contract_deployers = {
        "safe_V0_0_1": SafeV001.deploy_contract,
        "safe_V1_0_0": SafeV100.deploy_contract,
        "safe_V1_1_1": SafeV111.deploy_contract,
        "safe_V1_3_0": SafeV130.deploy_contract,
        "safe_V1_4_1": SafeV141.deploy_contract,
        "compatibility_fallback_handler": Safe.deploy_compatibility_fallback_handler,
        "simulate_tx_accessor_V1_4_1": Safe.deploy_simulate_tx_accessor,
        "proxy_factory": ProxyFactoryV141.deploy_contract,
        "multi_send": MultiSend.deploy_contract,
    }

    contract_addresses: Dict[str, ChecksumAddress] = {}

    @property
    def safe_contract(self):
        """
        :return: Last Safe Contract available
        """
        return self.safe_contract_V1_4_1

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.deploy_safe_singleton_factory()

        if not cls.contract_addresses:
            # First time mixin is called, deploy Safe contracts
            for key, function in cls.contract_deployers.items():
                try:
                    cls.contract_addresses[key] = function(
                        cls.ethereum_client, cls.ethereum_test_account
                    ).contract_address
                except ContractAlreadyDeployed as e:
                    cls.contract_addresses[key] = e.address

        cls.configure_django_settings()
        cls.configure_envvars()

        cls.compatibility_fallback_handler = (
            get_compatibility_fallback_handler_contract(
                cls.w3, cls.contract_addresses["compatibility_fallback_handler"]
            )
        )
        cls.simulate_tx_accessor_V1_4_1 = get_simulate_tx_accessor_V1_4_1_contract(
            cls.w3, cls.contract_addresses["simulate_tx_accessor_V1_4_1"]
        )
        cls.safe_contract_V1_4_1 = get_safe_V1_4_1_contract(
            cls.w3, cls.contract_addresses["safe_V1_4_1"]
        )
        cls.safe_contract_V1_3_0 = get_safe_V1_3_0_contract(
            cls.w3, cls.contract_addresses["safe_V1_3_0"]
        )
        cls.safe_contract_V1_1_1 = get_safe_V1_1_1_contract(
            cls.w3, cls.contract_addresses["safe_V1_1_1"]
        )
        cls.safe_contract_V1_0_0 = get_safe_V1_0_0_contract(
            cls.w3, cls.contract_addresses["safe_V1_0_0"]
        )
        cls.safe_contract_V0_0_1 = get_safe_V1_0_0_contract(
            cls.w3, cls.contract_addresses["safe_V0_0_1"]
        )
        cls.proxy_factory_contract = get_proxy_factory_contract(
            cls.w3, cls.contract_addresses["proxy_factory"]
        )
        cls.proxy_factory = ProxyFactory(
            cls.proxy_factory_contract.address, cls.ethereum_client
        )
        cls.multi_send_contract = get_multi_send_contract(
            cls.w3, cls.contract_addresses["multi_send"]
        )
        cls.multi_send = MultiSend(
            cls.ethereum_client, address=cls.multi_send_contract.address
        )

    @classmethod
    def configure_django_settings(cls):
        """
        Configure settings for django based applications

        :return:
        """

        try:
            from django.conf import settings

            settings.SAFE_CONTRACT_ADDRESS = cls.contract_addresses["safe_V1_4_1"]
            settings.SAFE_DEFAULT_CALLBACK_HANDLER = cls.contract_addresses[
                "compatibility_fallback_handler"
            ]
            settings.SAFE_MULTISEND_ADDRESS = cls.contract_addresses["multi_send"]
            settings.SAFE_PROXY_FACTORY_ADDRESS = cls.contract_addresses[
                "proxy_factory"
            ]
            settings.SAFE_V0_0_1_CONTRACT_ADDRESS = cls.contract_addresses[
                "safe_V0_0_1"
            ]
            settings.SAFE_V1_0_0_CONTRACT_ADDRESS = cls.contract_addresses[
                "safe_V1_0_0"
            ]
            settings.SAFE_V1_1_1_CONTRACT_ADDRESS = cls.contract_addresses[
                "safe_V1_1_1"
            ]
            settings.SAFE_V1_3_0_CONTRACT_ADDRESS = cls.contract_addresses[
                "safe_V1_3_0"
            ]
            settings.SAFE_V1_4_1_CONTRACT_ADDRESS = cls.contract_addresses[
                "safe_V1_4_1"
            ]
            settings.SAFE_SIMULATE_TX_ACCESSOR = cls.contract_addresses[
                "simulate_tx_accessor_V1_4_1"
            ]
            settings.SAFE_VALID_CONTRACT_ADDRESSES = {
                settings.SAFE_CONTRACT_ADDRESS,
                settings.SAFE_V1_3_0_CONTRACT_ADDRESS,
                settings.SAFE_V1_1_1_CONTRACT_ADDRESS,
                settings.SAFE_V1_0_0_CONTRACT_ADDRESS,
                settings.SAFE_V0_0_1_CONTRACT_ADDRESS,
            }

        except ModuleNotFoundError:
            logger.info("Django library is not installed")

    @classmethod
    def configure_envvars(cls):
        """
        Configure environment variables

        :return:
        """
        os.environ["SAFE_SIMULATE_TX_ACCESSOR_ADDRESS"] = cls.contract_addresses[
            "simulate_tx_accessor_V1_4_1"
        ]

    @classmethod
    def deploy_safe_singleton_factory(cls) -> bool:
        """
        Deploy Safe Singleton Factory for deterministic deployments and speeding up tests,
        due to being able to check quickly if contracts are deployed in their expected addresses

        Singleton factory with `1337` chainId is used
        Deployer address: 0xE1CB04A0fA36DdD16a06ea828007E35e1a3cBC37
        Expected factory address: 0x914d7Fec6aaC8cd542e72Bca78B30650d45643d7

        :return: `True` if deployed, `False` otherwise
        """
        if cls.ethereum_client.get_singleton_factory_address():
            return False

        raw_tx = HexBytes(
            "0xf8a78085174876e800830186a08080b853604580600e600039806000f350fe7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffe03601600081602082378035828234f58015156039578182fd5b8082525050506014600cf3820a96a0460c6ea9b8f791e5d9e67fbf2c70aba92bf88591c39ac3747ea1bedc2ef1750ca04b08a4b5cea15a56276513da7a0c0b34f16e89811d5dd911efba5f8625a921cc"
        )
        send_tx(
            cls.w3,
            {"to": SAFE_SINGLETON_FACTORY_DEPLOYER_ADDRESS, "value": 10000000000000000},
            cls.ethereum_test_account,
        )
        tx_hash = cls.ethereum_client.send_raw_transaction(raw_tx)
        tx_receipt = cls.ethereum_client.get_transaction_receipt(tx_hash, timeout=30)
        assert tx_receipt["status"] == 1

        # Clear cached empty singleton factory
        cls.ethereum_client.get_singleton_factory_address.cache_clear()

        return True

    def deploy_test_safe(self, *args, **kwargs) -> Safe:
        """
        :param args:
        :param kwargs:
        :return: Deploy last available Safe
        """
        return self.deploy_test_safe_v1_4_1(*args, **kwargs)

    def _deploy_test_safe(
        self,
        initializer: bytes,
        master_copy_address: ChecksumAddress,
        initial_funding_wei: Optional[Wei] = None,
    ) -> Safe:
        """
        Internal method to deploy a Safe given the initializer and master copy

        :param initializer:
        :param master_copy_address:
        :param initial_funding_wei: If provided, funds will be sent to the Safe
        :return: A deployed Safe
        """
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            self.ethereum_test_account,
            master_copy_address,
            initializer=initializer,
        )
        safe = Safe(
            ethereum_tx_sent.contract_address,
            self.ethereum_client,
            simulate_tx_accessor_address=self.simulate_tx_accessor_V1_4_1.address,
        )

        if initial_funding_wei:
            self.send_ether(safe.address, initial_funding_wei)

        return safe

    def _deploy_new_test_safe(
        self,
        master_copy_version: str,
        master_copy_address: ChecksumAddress,
        number_owners: int = 3,
        threshold: Optional[int] = None,
        owners: Optional[List[ChecksumAddress]] = None,
        initial_funding_wei: int = 0,
        fallback_handler: Optional[ChecksumAddress] = None,
    ) -> Safe:
        """
        Internal method to deploy Safes from 1.1.1 to 1.4.1, as setup method didn't change

        :param master_copy_version:
        :param master_copy_address:
        :param number_owners:
        :param threshold:
        :param owners:
        :param initial_funding_wei:
        :param fallback_handler:
        :return: A deployed Safe
        """

        fallback_handler = (
            fallback_handler or self.compatibility_fallback_handler.address
        )
        owners = (
            owners
            if owners
            else [Account.create().address for _ in range(number_owners)]
        )
        if not threshold:
            threshold = len(owners) - 1 if len(owners) > 1 else 1
        to = NULL_ADDRESS
        data = b""
        payment_token = NULL_ADDRESS
        payment = 0
        payment_receiver = NULL_ADDRESS
        initializer = HexBytes(
            self.safe_contract_V1_4_1.functions.setup(
                owners,
                threshold,
                to,
                data,
                fallback_handler,
                payment_token,
                payment,
                payment_receiver,
            ).build_transaction(get_empty_tx_params())["data"]
        )

        safe = self._deploy_test_safe(
            initializer, master_copy_address, initial_funding_wei=initial_funding_wei
        )

        self.assertEqual(safe.retrieve_version(), master_copy_version)
        self.assertEqual(safe.retrieve_threshold(), threshold)
        self.assertCountEqual(safe.retrieve_owners(), owners)

        return safe

    def deploy_test_safe_v1_4_1(
        self,
        number_owners: int = 3,
        threshold: Optional[int] = None,
        owners: Optional[List[ChecksumAddress]] = None,
        initial_funding_wei: int = 0,
        fallback_handler: Optional[ChecksumAddress] = None,
    ) -> Safe:
        """
        Deploy a Safe v1.4.1

        :param number_owners:
        :param threshold:
        :param owners:
        :param initial_funding_wei:
        :param fallback_handler:
        :return:
        """
        return self._deploy_new_test_safe(
            "1.4.1",
            self.safe_contract_V1_4_1.address,
            number_owners=number_owners,
            threshold=threshold,
            owners=owners,
            initial_funding_wei=initial_funding_wei,
            fallback_handler=fallback_handler,
        )

    def deploy_test_safe_v1_3_0(
        self,
        number_owners: int = 3,
        threshold: Optional[int] = None,
        owners: Optional[List[ChecksumAddress]] = None,
        initial_funding_wei: int = 0,
        fallback_handler: Optional[ChecksumAddress] = None,
    ) -> Safe:
        """
        Deploy a Safe v1.3.0

        :param number_owners:
        :param threshold:
        :param owners:
        :param initial_funding_wei:
        :param fallback_handler:
        :return:
        """
        return self._deploy_new_test_safe(
            "1.3.0",
            self.safe_contract_V1_3_0.address,
            number_owners=number_owners,
            threshold=threshold,
            owners=owners,
            initial_funding_wei=initial_funding_wei,
            fallback_handler=fallback_handler,
        )

    def deploy_test_safe_v1_1_1(
        self,
        number_owners: int = 3,
        threshold: Optional[int] = None,
        owners: Optional[List[ChecksumAddress]] = None,
        initial_funding_wei: int = 0,
        fallback_handler: Optional[ChecksumAddress] = NULL_ADDRESS,
    ) -> Safe:
        return self._deploy_new_test_safe(
            "1.1.1",
            self.safe_contract_V1_1_1.address,
            number_owners=number_owners,
            threshold=threshold,
            owners=owners,
            initial_funding_wei=initial_funding_wei,
            fallback_handler=fallback_handler,
        )

    def deploy_test_safe_v1_0_0(
        self,
        number_owners: int = 3,
        threshold: Optional[int] = None,
        owners: Optional[List[ChecksumAddress]] = None,
        initial_funding_wei: int = 0,
    ) -> Safe:
        owners = (
            owners
            if owners
            else [Account.create().address for _ in range(number_owners)]
        )
        if not threshold:
            threshold = len(owners) - 1 if len(owners) > 1 else 1
        to = NULL_ADDRESS
        data = b""
        payment_token = NULL_ADDRESS
        payment = 0
        payment_receiver = NULL_ADDRESS
        initializer = HexBytes(
            self.safe_contract_V1_0_0.functions.setup(
                owners, threshold, to, data, payment_token, payment, payment_receiver
            ).build_transaction(get_empty_tx_params())["data"]
        )
        safe = self._deploy_test_safe(
            initializer,
            self.safe_contract_V1_0_0.address,
            initial_funding_wei=initial_funding_wei,
        )

        self.assertEqual(safe.retrieve_version(), "1.0.0")
        self.assertEqual(safe.retrieve_threshold(), threshold)
        self.assertCountEqual(safe.retrieve_owners(), owners)

        return safe

    def deploy_example_guard(self) -> ChecksumAddress:
        """
        :return: An example DebugTransactionGuard (from safe contracts v1.4.1) supporting IERC165
        """

        bytecode = " "
        guard_contract = self.w3.eth.contract(abi=[], bytecode=bytecode)
        tx_hash = guard_contract.constructor().transact(
            {"from": self.w3.eth.accounts[0]}
        )
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        assert tx_receipt["status"] == 1, "Problem deploying example guard"
        return tx_receipt["contractAddress"]

    def deploy_sign_message_lib(self) -> ChecksumAddress:
        """
        Deploy sign message lib

        :return: Deployed address
        """
        contract = get_sign_message_lib_contract(self.ethereum_client.w3)
        tx_hash = contract.constructor().transact({"from": self.w3.eth.accounts[0]})
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        assert tx_receipt["status"] == 1, "Problem deploying sign message lib"
        return tx_receipt["contractAddress"]



import dataclasses
from typing import Any, Dict, List, Optional

from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3.types import LogReceipt

from gnosis.eth.account_abstraction.constants import (
    ACCOUNT_DEPLOYED_TOPIC,
    DEPOSIT_EVENT_TOPIC,
    EXECUTION_FROM_MODULE_FAILURE_TOPIC,
    EXECUTION_FROM_MODULE_SUCCESS_TOPIC,
)
from gnosis.eth.utils import fast_to_checksum_address


@dataclasses.dataclass(eq=True, frozen=True)
class UserOperationReceipt:
    user_operation_hash: bytes
    entry_point: ChecksumAddress
    sender: ChecksumAddress
    nonce: int
    paymaster: ChecksumAddress
    actual_gas_cost: int
    actual_gas_used: int
    success: bool
    reason: Optional[str]
    logs: List[LogReceipt]

    @classmethod
    def from_bundler_response(
        cls,
        user_operation_receipt_response: Dict[str, Any],
    ) -> "UserOperationReceipt":
        return cls(
            HexBytes(user_operation_receipt_response["userOpHash"]),
            user_operation_receipt_response["entryPoint"],
            user_operation_receipt_response["sender"],
            int(user_operation_receipt_response["nonce"], 16),
            user_operation_receipt_response["paymaster"],
            int(user_operation_receipt_response["actualGasCost"], 16),
            int(user_operation_receipt_response["actualGasUsed"], 16),
            user_operation_receipt_response["success"],
            user_operation_receipt_response.get("reason"),
            user_operation_receipt_response["logs"],
        )

    def get_deployed_account(self) -> Optional[ChecksumAddress]:
        """
        :return: Deployed account in case a new account was deployed
        """
        for log in self.logs:
            if (
                len(log["topics"]) == 3
                and HexBytes(log["topics"][0]) == ACCOUNT_DEPLOYED_TOPIC
            ):
                # Address is stored at the 40 last chars of the 3rd topic
                return fast_to_checksum_address(log["topics"][2][-40:])
        return None

    def get_deposit(self) -> int:
        """
        :return: Deposited value on the entrypoint for running the UserOperationReceipt
        """
        deposited = 0
        for log in self.logs:
            if (
                len(log["topics"]) == 2
                and HexBytes(log["topics"][0]) == DEPOSIT_EVENT_TOPIC
                and fast_to_checksum_address(log["address"]) == self.entry_point
                and fast_to_checksum_address(log["topics"][1][-40:]) == self.sender
            ):
                deposited += int(log["data"], 16)
        return deposited

    def get_module_address(self) -> Optional[ChecksumAddress]:
        """
        Use Safe's `ExecutionFromModuleSuccess` event to get the 4337 module address

        :return: If using a ``Safe``, the ``4337 module address`` used, ``None`` otherwise
        """
        for log in reversed(self.logs):
            if (
                len(log["topics"]) == 2
                and HexBytes(log["topics"][0])
                in (
                    EXECUTION_FROM_MODULE_SUCCESS_TOPIC,
                    EXECUTION_FROM_MODULE_FAILURE_TOPIC,
                )
                and fast_to_checksum_address(log["address"]) == self.sender
            ):
                return fast_to_checksum_address(log["topics"][1][-40:])
        return None



from typing import Any, List, Optional, TypedDict

from eth_typing import AnyAddress, HexStr


class ParameterDecoded(TypedDict):
    name: str
    type: str
    value: Any


class DataDecoded(TypedDict):
    method: str
    parameters: List[ParameterDecoded]


class Erc20Info(TypedDict):
    name: str
    symbol: str
    decimals: int
    logo_uri: str


class Balance(TypedDict):
    token_address: Optional[AnyAddress]
    token: Optional[Erc20Info]
    balance: int


class DelegateUser(TypedDict):
    safe: Optional[AnyAddress]
    delegate: AnyAddress
    delegator: AnyAddress
    label: str


class MessageConfirmation(TypedDict):
    created: str
    modified: str
    owner: AnyAddress
    signature: HexStr
    signatureType: str


class Message(TypedDict):
    created: str
    modified: str
    safe: AnyAddress
    messageHash: HexStr
    message: Any
    proposedBy: AnyAddress
    safeAppId: int
    confirmations: Optional[List[MessageConfirmation]]
    preparedSignature: Optional[HexStr]


class TransactionConfirmation(TypedDict):
    owner: AnyAddress
    submissionDate: str
    transactionHash: HexStr
    signature: HexStr
    signatureType: str


class Transaction(TypedDict):
    safe: AnyAddress
    to: AnyAddress
    value: str
    data: Optional[HexStr]
    operation: int
    gasToken: Optional[AnyAddress]
    safeTxGas: int
    baseGas: int
    gasPrice: str
    refundReceiver: Optional[AnyAddress]
    nonce: int
    execution_date: str
    submission_date: str
    modified: str
    blockNumber: Optional[int]
    transactionHash: HexStr
    safeTxHash: HexStr
    proposer: AnyAddress
    executor: Optional[AnyAddress]
    isExecuted: bool
    isSuccessful: Optional[bool]
    ethGasPrice: Optional[str]
    maxFeePerGas: Optional[str]
    maxPriorityFeePerGas: Optional[str]
    gasUsed: Optional[int]
    fee: Optional[int]
    origin: Optional[str]
    dataDecoded: Optional[List[DataDecoded]]
    confirmationsRequired: int
    confirmations: Optional[List[TransactionConfirmation]]
    trusted: bool
    signatures: Optional[HexStr]



import dataclasses
import datetime
import zoneinfo
from unittest import TestCase

from gnosis.eth.account_abstraction import UserOperation
from gnosis.eth.tests.mocks.mock_bundler import (
    safe_4337_chain_id_mock,
    safe_4337_module_address_mock,
    safe_4337_module_domain_separator_mock,
    safe_4337_safe_operation_hash_mock,
    safe_4337_user_operation_hash_mock,
    user_operation_mock,
    user_operation_with_valid_dates_hash_mock,
    user_operation_with_valid_dates_mock,
)

from ...account_abstraction import SafeOperation
from ...account_abstraction.safe_operation import _domain_separator_cache


class TestSafeOperation(TestCase):
    def setUp(self):
        _domain_separator_cache.clear()

    def tearDown(self):
        _domain_separator_cache.clear()

    def test_from_user_operation(self):
        safe_operation = SafeOperation.from_user_operation(
            UserOperation.from_bundler_response(
                safe_4337_user_operation_hash_mock, user_operation_mock["result"]
            )
        )

        self.assertDictEqual(_domain_separator_cache, {})

        self.assertEqual(
            safe_operation.get_domain_separator(
                safe_4337_chain_id_mock, safe_4337_module_address_mock
            ),
            safe_4337_module_domain_separator_mock,
        )
        self.assertDictEqual(
            _domain_separator_cache,
            {
                (
                    safe_4337_chain_id_mock,
                    safe_4337_module_address_mock,
                ): safe_4337_module_domain_separator_mock
            },
        )

        self.assertEqual(
            safe_operation.get_safe_operation_hash(
                safe_4337_chain_id_mock, safe_4337_module_address_mock
            ),
            safe_4337_safe_operation_hash_mock,
        )
        self.assertDictEqual(
            _domain_separator_cache,
            {
                (
                    safe_4337_chain_id_mock,
                    safe_4337_module_address_mock,
                ): safe_4337_module_domain_separator_mock
            },
        )

        self.assertIsNone(safe_operation.valid_after_as_datetime)
        self.assertIsNone(safe_operation.valid_until_as_datetime)

    def test_datetime_parse(self):
        safe_operation = SafeOperation.from_user_operation(
            UserOperation.from_bundler_response(
                user_operation_with_valid_dates_hash_mock,
                user_operation_with_valid_dates_mock["result"],
            )
        )

        self.assertEqual(safe_operation.valid_after, 1710848424)
        self.assertEqual(
            safe_operation.valid_after_as_datetime,
            datetime.datetime(
                2024, 3, 19, 11, 40, 24, tzinfo=zoneinfo.ZoneInfo(key="UTC")
            ),
        )
        self.assertEqual(safe_operation.valid_until, 1710908424)
        self.assertEqual(
            safe_operation.valid_until_as_datetime,
            datetime.datetime(
                2024, 3, 20, 4, 20, 24, tzinfo=zoneinfo.ZoneInfo(key="UTC")
            ),
        )

        # Test invalid value cannot be parsed as datetime
        invalid_safe_operation = dataclasses.replace(
            safe_operation,
            valid_after=5555555555555555555555,
            valid_until=666666666666666666666,
        )
        self.assertIsNone(invalid_safe_operation.valid_after_as_datetime)
        self.assertIsNone(invalid_safe_operation.valid_until_as_datetime)



import logging
from functools import cached_property
from typing import Optional

from eth_abi.exceptions import DecodingError
from web3.exceptions import Web3Exception

from .. import EthereumClient, EthereumNetwork
from ..contracts import get_kyber_network_proxy_contract
from .exceptions import CannotGetPriceFromOracle, InvalidPriceFromOracle
from .oracles import PriceOracle
from .utils import get_decimals

logger = logging.getLogger(__name__)


class KyberOracle(PriceOracle):
    """
    KyberSwap Legacy Oracle

    https://docs.kyberswap.com/Legacy/addresses/addresses-mainnet
    """

    # This is the `tokenAddress` they use for ETH ¯\_(ツ)_/¯
    ETH_TOKEN_ADDRESS = "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"
    ADDRESSES = {
        EthereumNetwork.MAINNET: "0x9AAb3f75489902f3a48495025729a0AF77d4b11e",
        EthereumNetwork.RINKEBY: "0x0d5371e5EE23dec7DF251A8957279629aa79E9C5",
        EthereumNetwork.ROPSTEN: "0xd719c34261e099Fdb33030ac8909d5788D3039C4",
    }

    def __init__(
        self,
        ethereum_client: EthereumClient,
        kyber_network_proxy_address: Optional[str] = None,
    ):
        """
        :param ethereum_client:
        :param kyber_network_proxy_address: https://developer.kyber.network/docs/MainnetEnvGuide/#contract-addresses
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self._kyber_network_proxy_address = kyber_network_proxy_address

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        return ethereum_client.get_network() in cls.ADDRESSES

    @cached_property
    def kyber_network_proxy_address(self):
        if self._kyber_network_proxy_address:
            return self._kyber_network_proxy_address
        return self.ADDRESSES.get(
            self.ethereum_client.get_network(),
            self.ADDRESSES.get(EthereumNetwork.MAINNET),
        )  # By default return Mainnet address

    @cached_property
    def kyber_network_proxy_contract(self):
        return get_kyber_network_proxy_contract(
            self.w3, self.kyber_network_proxy_address
        )

    def get_price(
        self, token_address_1: str, token_address_2: str = ETH_TOKEN_ADDRESS
    ) -> float:
        if token_address_1 == token_address_2:
            return 1.0
        try:
            # Get decimals for token, estimation will be more accurate
            decimals = get_decimals(token_address_1, self.ethereum_client)
            token_unit = int(10**decimals)
            (
                expected_rate,
                _,
            ) = self.kyber_network_proxy_contract.functions.getExpectedRate(
                token_address_1, token_address_2, int(token_unit)
            ).call()

            price = expected_rate / 1e18

            if price <= 0.0:
                # Try again the opposite
                (
                    expected_rate,
                    _,
                ) = self.kyber_network_proxy_contract.functions.getExpectedRate(
                    token_address_2, token_address_1, int(token_unit)
                ).call()
                price = (token_unit / expected_rate) if expected_rate else 0

            if price <= 0.0:
                message = (
                    f"price={price} <= 0 from kyber-network-proxy={self.kyber_network_proxy_address} "
                    f"for token-1={token_address_1} to token-2={token_address_2}"
                )
                logger.debug(message)
                raise InvalidPriceFromOracle(message)
            return price
        except (Web3Exception, DecodingError, ValueError) as e:
            message = (
                f"Cannot get price from kyber-network-proxy={self.kyber_network_proxy_address} "
                f"for token-1={token_address_1} to token-2={token_address_2}"
            )
            logger.debug(message)
            raise CannotGetPriceFromOracle(message) from e



from eth_typing import ChecksumAddress

from gnosis.safe import SafeTx


class TransactionServiceTx(SafeTx):
    def __init__(self, proposer: ChecksumAddress, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.proposer = proposer



import json
from functools import cached_property
from typing import Any, Dict, List, Optional, TypedDict, Union, cast

from eth_account import Account
from eth_account.messages import encode_defunct
from eth_typing import AnyAddress, ChecksumAddress, HexStr

from gnosis.eth import EthereumNetwork, EthereumNetworkNotSupported
from gnosis.eth.eip712 import eip712_encode_hash

from ..eth.constants import NULL_ADDRESS
from ..util.http import prepare_http_session
from .order import Order, OrderKind


class TradeResponse(TypedDict):
    blockNumber: int
    logIndex: int
    orderUid: HexStr
    buyAmount: str  # Stringified int
    sellAmount: str  # Stringified int
    sellAmountBeforeFees: str  # Stringified int
    owner: AnyAddress  # Not checksummed
    buyToken: AnyAddress
    sellToken: AnyAddress
    txHash: HexStr


class AmountResponse(TypedDict):
    sellAmount: int
    buyAmount: int


class ErrorResponse(TypedDict):
    errorType: str
    description: str


class CowSwapAPI:
    """
    Client for CowSwap API. More info: https://docs.cowswap.exchange/
    """

    SETTLEMENT_CONTRACT_ADDRESSES = {
        EthereumNetwork.MAINNET: "0x9008D19f58AAbD9eD0D60971565AA8510560ab41",
        EthereumNetwork.GNOSIS: "0x9008D19f58AAbD9eD0D60971565AA8510560ab41",
        EthereumNetwork.SEPOLIA: "0x9008D19f58AAbD9eD0D60971565AA8510560ab41",
    }

    API_BASE_URLS = {
        EthereumNetwork.MAINNET: "https://api.cow.fi/mainnet",
        EthereumNetwork.GNOSIS: "https://api.cow.fi/xdai",
        EthereumNetwork.SEPOLIA: "https://api.cow.fi/sepolia",
    }

    def __init__(self, ethereum_network: EthereumNetwork, request_timeout: int = 10):
        self.network = ethereum_network
        if self.network not in self.API_BASE_URLS:
            raise EthereumNetworkNotSupported(
                f"{self.network.name} network not supported by CowSwap"
            )
        self.settlement_contract_address = self.SETTLEMENT_CONTRACT_ADDRESSES[
            self.network
        ]
        self.base_url = self.API_BASE_URLS[self.network]
        self.http_session = prepare_http_session(10, 100)
        self.request_timeout = request_timeout

    @cached_property
    def weth_address(self) -> ChecksumAddress:
        """
        :return: Wrapped ether checksummed address
        """
        if self.network == EthereumNetwork.GNOSIS:  # WXDAI
            return ChecksumAddress("0xe91D153E0b41518A2Ce8Dd3D7944Fa863463a97d")

        # Mainnet WETH9
        return ChecksumAddress("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2")

    def get_quote(
        self, order: Order, from_address: ChecksumAddress
    ) -> Union[Dict[str, Any], ErrorResponse]:
        url = self.base_url + "/api/v1/quote"
        data_json = {
            "sellToken": order.sellToken.lower(),
            "buyToken": order.buyToken.lower(),
            "sellAmountAfterFee": str(order.sellAmount),
            # "validTo": order.validTo,
            "appData": json.dumps(order.appData),
            "feeAmount": str(order.feeAmount),
            "kind": order.kind,
            "partiallyFillable": order.partiallyFillable,
            "signingScheme": "ethsign",
            "from": from_address,
            "priceQuality": "fast",
        }
        r = self.http_session.post(url, json=data_json, timeout=self.request_timeout)
        if r.ok:
            return r.json()
        else:
            return ErrorResponse(r.json())

    def get_fee(
        self, order: Order, from_address: ChecksumAddress
    ) -> Union[int, ErrorResponse]:
        quote = self.get_quote(order, from_address)

        if "quote" in quote:
            return int(quote["quote"]["feeAmount"])
        else:
            return quote

    def place_order(
        self, order: Order, private_key: HexStr
    ) -> Union[HexStr, ErrorResponse]:
        """
        Place order. If `feeAmount=0` in Order it will be calculated calling `get_fee(order, from_address)`

        :return: UUID for the order as a hex hash
        """
        assert (
            order.buyAmount and order.sellAmount
        ), "Order buyAmount and sellAmount cannot be empty"

        url = self.base_url + "/api/v1/orders"
        from_address = Account.from_key(private_key).address
        if not order.feeAmount:
            fee_amount = self.get_fee(order, from_address)
            if isinstance(fee_amount, int):
                order.feeAmount = fee_amount
            elif "errorType" in fee_amount:  # ErrorResponse
                return fee_amount

        signable_hash = eip712_encode_hash(
            order.get_eip712_structured_data(
                self.network.value, self.settlement_contract_address
            )
        )
        message = encode_defunct(primitive=signable_hash)
        signed_message = Account.from_key(private_key).sign_message(message)

        data_json = {
            "sellToken": order.sellToken.lower(),
            "buyToken": order.buyToken.lower(),
            "sellAmount": str(order.sellAmount),
            "buyAmount": str(order.buyAmount),
            "validTo": order.validTo,
            "appData": json.dumps(order.appData),
            "feeAmount": str(order.feeAmount),
            "kind": order.kind,
            "partiallyFillable": order.partiallyFillable,
            "signature": signed_message.signature.hex(),
            "signingScheme": "ethsign",
            "from": from_address,
        }
        r = self.http_session.post(url, json=data_json, timeout=self.request_timeout)
        if r.ok:
            return HexStr(r.json())
        else:
            return ErrorResponse(r.json())

    def get_orders(
        self, owner: ChecksumAddress, offset: int = 0, limit=10
    ) -> List[Dict[str, Any]]:
        """
        :param owner:
        :param offset: Defaults to 0
        :param limit: Defaults to 10. Maximum is 1000, minimum is 1
        :return: Orders of one user paginated. The orders are ordered by their creation
            date descending (newest orders first).
            To enumerate all orders start with offset 0 and keep increasing the offset by the
            total number of returned results. When a response contains less than the limit
            the last page has been reached.
        """
        url = self.base_url + f"/api/v1/account/{owner}/orders"
        r = self.http_session.get(url, timeout=self.request_timeout)
        if r.ok:
            return cast(List[Dict[str, Any]], r.json())
        else:
            return ErrorResponse(r.json())

    def get_trades(
        self, order_ui: Optional[HexStr] = None, owner: Optional[ChecksumAddress] = None
    ) -> List[TradeResponse]:
        assert bool(order_ui) ^ bool(
            owner
        ), "order_ui or owner must be provided, but not both"
        url = self.base_url + "/api/v1/trades/?"
        if order_ui:
            url += f"orderUid={order_ui}"
        elif owner:
            url += f"owner={owner}"

        r = self.http_session.get(url, timeout=self.request_timeout)
        if r.ok:
            return cast(List[TradeResponse], r.json())
        else:
            return ErrorResponse(r.json())

    def get_estimated_amount(
        self,
        base_token: ChecksumAddress,
        quote_token: ChecksumAddress,
        kind: OrderKind,
        amount_wei: int,
    ) -> Union[AmountResponse, ErrorResponse]:
        """

        :param base_token:
        :param quote_token:
        :param kind:
        :param amount_wei:
        :return: Both `sellAmount` and `buyAmount` as they can be adjusted by CowSwap API
        """
        order = Order(
            sellToken=base_token,
            buyToken=quote_token,
            receiver=NULL_ADDRESS,
            sellAmount=amount_wei * 10 if kind == OrderKind.SELL else 0,
            buyAmount=amount_wei * 10 if kind == OrderKind.BUY else 0,
            validTo=0,  # Valid for 1 hour
            appData={},
            feeAmount=0,
            kind=kind.name.lower(),  # `sell` or `buy`
            partiallyFillable=False,
            sellTokenBalance="erc20",  # `erc20`, `external` or `internal`
            buyTokenBalance="erc20",  # `erc20` or `internal`
        )

        quote = self.get_quote(order, NULL_ADDRESS)
        if "quote" in quote:
            return {
                "buyAmount": int(quote["quote"]["buyAmount"]),
                "sellAmount": int(quote["quote"]["sellAmount"]),
            }
        else:
            return quote



from unittest import TestCase

from ...account_abstraction import UserOperation
from ...account_abstraction.user_operation import UserOperationV07
from ..mocks.mock_bundler import (
    safe_4337_chain_id_mock,
    safe_4337_user_operation_hash_mock,
    user_operation_mock,
    user_operation_v07_hash,
    user_operation_v07_mock,
)


class TestUserOperation(TestCase):
    def test_calculate_user_operation_hash_V06(self):
        user_operation_hash = safe_4337_user_operation_hash_mock
        user_operation = UserOperation.from_bundler_response(
            user_operation_hash.hex(), user_operation_mock["result"]
        )
        self.assertIsInstance(user_operation, UserOperation)
        self.assertEqual(
            user_operation.calculate_user_operation_hash(safe_4337_chain_id_mock),
            user_operation_hash,
        )

    def test_calculate_user_operation_hash_V07(self):
        user_operation_hash = user_operation_v07_hash
        user_operation = UserOperation.from_bundler_response(
            user_operation_hash.hex(), user_operation_v07_mock["result"]
        )
        self.assertIsInstance(user_operation, UserOperationV07)
        self.assertEqual(
            user_operation.calculate_user_operation_hash(safe_4337_chain_id_mock),
            user_operation_hash,
        )



import secrets
from abc import ABCMeta
from functools import cache
from typing import Callable, Optional

from eth_abi.packed import encode_packed
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract.contract import Contract, ContractFunction

from gnosis.eth import EthereumClient, EthereumTxSent
from gnosis.eth.contracts import (
    ContractBase,
    get_paying_proxy_deployed_bytecode,
    get_proxy_1_0_0_deployed_bytecode,
    get_proxy_1_1_1_deployed_bytecode,
    get_proxy_1_1_1_mainnet_deployed_bytecode,
    get_proxy_1_3_0_deployed_bytecode,
    get_proxy_1_4_1_deployed_bytecode,
    get_proxy_factory_V1_0_0_contract,
    get_proxy_factory_V1_1_1_contract,
    get_proxy_factory_V1_3_0_contract,
    get_proxy_factory_V1_4_1_contract,
)
from gnosis.eth.utils import (
    compare_byte_code,
    fast_keccak,
    get_empty_tx_params,
    mk_contract_address_2,
)


class ProxyFactory(ContractBase, metaclass=ABCMeta):
    def __new__(cls, *args, version: str = "1.4.1", **kwargs) -> "ProxyFactory":
        if cls is not ProxyFactory:
            return super().__new__(cls)

        versions = {
            "1.0.0": ProxyFactoryV100,
            "1.1.1": ProxyFactoryV111,
            "1.3.0": ProxyFactoryV130,
            "1.4.1": ProxyFactoryV141,
        }
        instance_class = versions[version]
        instance = super().__new__(instance_class)
        return instance

    @classmethod
    def deploy_contract(
        cls, ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy Proxy Factory contract

        :param ethereum_client:
        :param deployer_account: Ethereum Account
        :return: ``EthereumTxSent`` with the deployed contract address
        """
        contract_fn = cls.get_contract_fn(cls)
        contract = contract_fn(ethereum_client.w3)
        constructor_data = contract.constructor().build_transaction(
            get_empty_tx_params()
        )["data"]
        return ethereum_client.deploy_and_initialize_contract(
            deployer_account, constructor_data
        )

    @cache
    def get_proxy_creation_code(self) -> bytes:
        """
        :return: Creation code used for the Proxy deployment.
            With this it is easily possible to calculate predicted address.
        """
        return self.contract.functions.proxyCreationCode().call()

    @cache
    def get_proxy_runtime_code(self) -> bytes:
        """
        :return: Runtime code of a deployed Proxy. For v1.4.1 onwards the method is not available, so `None`
            will be returned
        """
        return self.contract.functions.proxyRuntimeCode().call()

    def get_deploy_function(self, chain_specific: bool) -> ContractFunction:
        if chain_specific:
            raise NotImplementedError(
                f"createChainSpecificProxyWithNonce is not supported in {self.__class__.__name__}"
            )

        return self.contract.functions.createProxyWithNonce

    def check_proxy_code(self, address: ChecksumAddress) -> bool:
        """
        Check if proxy bytecode matches any of the deployed by the supported Proxy Factories

        :param address: Ethereum address to check
        :return: ``True`` if proxy is valid, ``False`` otherwise
        """

        def get_proxy_runtime_code() -> Optional[bytes]:
            try:
                return self.get_proxy_runtime_code()
            except NotImplementedError:
                return None

        deployed_proxy_code = self.w3.eth.get_code(address)
        proxy_code_fns = (
            get_proxy_1_4_1_deployed_bytecode,
            get_proxy_1_3_0_deployed_bytecode,
            get_proxy_1_1_1_deployed_bytecode,
            get_proxy_1_1_1_mainnet_deployed_bytecode,
            get_proxy_1_0_0_deployed_bytecode,
            get_paying_proxy_deployed_bytecode,
            get_proxy_runtime_code,
        )
        for proxy_code_fn in proxy_code_fns:
            proxy_code = proxy_code_fn()
            if proxy_code and compare_byte_code(deployed_proxy_code, proxy_code):
                return True
        return False

    def _deploy_proxy_contract(
        self,
        deployer_account: LocalAccount,
        deploy_fn: ContractFunction,
        gas: Optional[int] = None,
        gas_price: Optional[int] = None,
        nonce: Optional[int] = None,
    ) -> EthereumTxSent:
        """
        Common logic for `createProxy` and `createProxyWithNonce`

        :param deployer_account:
        :param deploy_fn:
        :param gas:
        :param gas_price:
        :param nonce:
        :return: EthereumTxSent
        """

        tx_params = self.ethereum_client.build_tx_params(
            from_address=deployer_account.address,
            gas=gas,
            gas_price=gas_price,
            nonce=nonce,
        )
        contract_address = deploy_fn.call(tx_params)
        tx = deploy_fn.build_transaction(tx_params)
        tx_hash = self.ethereum_client.send_unsigned_transaction(
            tx, private_key=deployer_account.key
        )

        return EthereumTxSent(tx_hash, tx, contract_address)

    def deploy_proxy_contract(
        self,
        deployer_account: LocalAccount,
        master_copy: ChecksumAddress,
        initializer: bytes = b"",
        gas: Optional[int] = None,
        gas_price: Optional[int] = None,
        nonce: Optional[int] = None,
    ) -> EthereumTxSent:
        """
        Deploy proxy contract via ProxyFactory using `createProxy` function (CREATE opcode)

        :param deployer_account: Ethereum account
        :param master_copy: Address the proxy will point at
        :param initializer: Initializer for the deployed proxy
        :param gas: Gas
        :param gas_price: Gas Price
        :param nonce: Nonce
        :return: EthereumTxSent
        """
        create_proxy_fn = self.contract.functions.createProxy(master_copy, initializer)

        return self._deploy_proxy_contract(
            deployer_account, create_proxy_fn, gas=gas, gas_price=gas_price, nonce=nonce
        )

    def calculate_proxy_address(
        self,
        master_copy: ChecksumAddress,
        initializer: bytes,
        salt_nonce: int,
        chain_specific: bool = False,
    ) -> ChecksumAddress:
        """
        Calculate proxy address for calling deploy_proxy_contract_with_nonce

        :param master_copy:
        :param initializer:
        :param salt_nonce:
        :param chain_specific: Calculate chain specific address (to prevent same address in other chains)
        :return:
        """

        if chain_specific:
            salt_nonce = fast_keccak(
                encode_packed(
                    ["bytes32", "uint256", "uint256"],
                    [
                        fast_keccak(initializer),
                        salt_nonce,
                        self.ethereum_client.get_chain_id(),
                    ],
                )
            )
        else:
            salt_nonce = fast_keccak(
                encode_packed(
                    ["bytes32", "uint256"], [fast_keccak(initializer), salt_nonce]
                )
            )

        proxy_creation_code = self.get_proxy_creation_code()
        deployment_data = encode_packed(
            ["bytes", "uint256"], [proxy_creation_code, int(master_copy, 0)]
        )

        return mk_contract_address_2(self.address, salt_nonce, deployment_data)

    def deploy_proxy_contract_with_nonce(
        self,
        deployer_account: LocalAccount,
        master_copy: ChecksumAddress,
        initializer: bytes = b"",
        salt_nonce: Optional[int] = None,
        gas: Optional[int] = None,
        gas_price: Optional[int] = None,
        nonce: Optional[int] = None,
        chain_specific: bool = False,
    ) -> EthereumTxSent:
        """
        Deploy proxy contract via Proxy Factory using `createProxyWithNonce` (CREATE2 opcode)

        :param deployer_account: Ethereum account
        :param master_copy: Address the proxy will point at
        :param initializer: Initializer for the deployed proxy
        :param salt_nonce: Uint256 for ``CREATE2`` salt. If not provided, a random one will be used
        :param gas: Gas
        :param gas_price: Gas Price
        :param nonce: Nonce
        :param chain_specific: Calculate chain specific address (to prevent same address in other chains)
        :return: EthereumTxSent
        """

        function = self.get_deploy_function(chain_specific)
        salt_nonce = salt_nonce if salt_nonce is not None else secrets.randbits(256)
        create_proxy_fn = function(master_copy, initializer, salt_nonce)

        return self._deploy_proxy_contract(
            deployer_account, create_proxy_fn, gas=gas, gas_price=gas_price, nonce=nonce
        )


class ProxyFactoryV100(ProxyFactory):
    def get_contract_fn(self) -> Callable[[Web3, ChecksumAddress], Contract]:
        return get_proxy_factory_V1_0_0_contract


class ProxyFactoryV111(ProxyFactory):
    def get_contract_fn(self) -> Callable[[Web3, ChecksumAddress], Contract]:
        return get_proxy_factory_V1_1_1_contract


class ProxyFactoryV130(ProxyFactory):
    def get_contract_fn(self) -> Callable[[Web3, ChecksumAddress], Contract]:
        return get_proxy_factory_V1_3_0_contract


class ProxyFactoryV141(ProxyFactory):
    def get_contract_fn(self) -> Callable[[Web3, ChecksumAddress], Contract]:
        return get_proxy_factory_V1_4_1_contract

    @cache
    def get_proxy_runtime_code(self) -> Optional[bytes]:
        """
        :return: From v1.4.1 onwards the method is not available
        :raises: NotImplementedError
        """
        raise NotImplementedError(
            "Deprecated, only creation code is available using `get_proxy_creation_code`"
        )

    def get_deploy_function(self, chain_specific: bool) -> ContractFunction:
        return (
            self.contract.functions.createChainSpecificProxyWithNonce
            if chain_specific
            else super().get_deploy_function(chain_specific)
        )

    def deploy_proxy_contract(self, *args, **kwargs):
        """
        .. deprecated:: ``createProxy`` function was deprecated in v1.4.1, use ``deploy_proxy_contract_with_nonce``

        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError("Deprecated, use `deploy_proxy_contract_with_nonce`")



balancer_pool_abi = [
    {
        "inputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "src",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "dst",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amt",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": True,
        "inputs": [
            {
                "indexed": True,
                "internalType": "bytes4",
                "name": "sig",
                "type": "bytes4",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "caller",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "bytes",
                "name": "data",
                "type": "bytes",
            },
        ],
        "name": "LOG_CALL",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "caller",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "tokenOut",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "tokenAmountOut",
                "type": "uint256",
            },
        ],
        "name": "LOG_EXIT",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "caller",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "tokenIn",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "tokenAmountIn",
                "type": "uint256",
            },
        ],
        "name": "LOG_JOIN",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "caller",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "tokenIn",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "tokenOut",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "tokenAmountIn",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "tokenAmountOut",
                "type": "uint256",
            },
        ],
        "name": "LOG_SWAP",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "src",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "dst",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amt",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "BONE",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "BPOW_PRECISION",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "EXIT_FEE",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "INIT_POOL_SUPPLY",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MAX_BOUND_TOKENS",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MAX_BPOW_BASE",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MAX_FEE",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MAX_IN_RATIO",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MAX_OUT_RATIO",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MAX_TOTAL_WEIGHT",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MAX_WEIGHT",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MIN_BALANCE",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MIN_BOUND_TOKENS",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MIN_BPOW_BASE",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MIN_FEE",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "MIN_WEIGHT",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "address", "name": "src", "type": "address"},
            {"internalType": "address", "name": "dst", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "amt", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "whom", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "balance", "type": "uint256"},
            {"internalType": "uint256", "name": "denorm", "type": "uint256"},
        ],
        "name": "bind",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "uint256", "name": "tokenBalanceIn", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightIn", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenBalanceOut", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightOut", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenAmountOut", "type": "uint256"},
            {"internalType": "uint256", "name": "swapFee", "type": "uint256"},
        ],
        "name": "calcInGivenOut",
        "outputs": [
            {"internalType": "uint256", "name": "tokenAmountIn", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "uint256", "name": "tokenBalanceIn", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightIn", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenBalanceOut", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightOut", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenAmountIn", "type": "uint256"},
            {"internalType": "uint256", "name": "swapFee", "type": "uint256"},
        ],
        "name": "calcOutGivenIn",
        "outputs": [
            {"internalType": "uint256", "name": "tokenAmountOut", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "uint256", "name": "tokenBalanceOut", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightOut", "type": "uint256"},
            {"internalType": "uint256", "name": "poolSupply", "type": "uint256"},
            {"internalType": "uint256", "name": "totalWeight", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenAmountOut", "type": "uint256"},
            {"internalType": "uint256", "name": "swapFee", "type": "uint256"},
        ],
        "name": "calcPoolInGivenSingleOut",
        "outputs": [
            {"internalType": "uint256", "name": "poolAmountIn", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "uint256", "name": "tokenBalanceIn", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightIn", "type": "uint256"},
            {"internalType": "uint256", "name": "poolSupply", "type": "uint256"},
            {"internalType": "uint256", "name": "totalWeight", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenAmountIn", "type": "uint256"},
            {"internalType": "uint256", "name": "swapFee", "type": "uint256"},
        ],
        "name": "calcPoolOutGivenSingleIn",
        "outputs": [
            {"internalType": "uint256", "name": "poolAmountOut", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "uint256", "name": "tokenBalanceIn", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightIn", "type": "uint256"},
            {"internalType": "uint256", "name": "poolSupply", "type": "uint256"},
            {"internalType": "uint256", "name": "totalWeight", "type": "uint256"},
            {"internalType": "uint256", "name": "poolAmountOut", "type": "uint256"},
            {"internalType": "uint256", "name": "swapFee", "type": "uint256"},
        ],
        "name": "calcSingleInGivenPoolOut",
        "outputs": [
            {"internalType": "uint256", "name": "tokenAmountIn", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "uint256", "name": "tokenBalanceOut", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightOut", "type": "uint256"},
            {"internalType": "uint256", "name": "poolSupply", "type": "uint256"},
            {"internalType": "uint256", "name": "totalWeight", "type": "uint256"},
            {"internalType": "uint256", "name": "poolAmountIn", "type": "uint256"},
            {"internalType": "uint256", "name": "swapFee", "type": "uint256"},
        ],
        "name": "calcSingleOutGivenPoolIn",
        "outputs": [
            {"internalType": "uint256", "name": "tokenAmountOut", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "uint256", "name": "tokenBalanceIn", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightIn", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenBalanceOut", "type": "uint256"},
            {"internalType": "uint256", "name": "tokenWeightOut", "type": "uint256"},
            {"internalType": "uint256", "name": "swapFee", "type": "uint256"},
        ],
        "name": "calcSpotPrice",
        "outputs": [
            {"internalType": "uint256", "name": "spotPrice", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "amt", "type": "uint256"},
        ],
        "name": "decreaseApproval",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "uint256", "name": "poolAmountIn", "type": "uint256"},
            {"internalType": "uint256[]", "name": "minAmountsOut", "type": "uint256[]"},
        ],
        "name": "exitPool",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "tokenOut", "type": "address"},
            {"internalType": "uint256", "name": "tokenAmountOut", "type": "uint256"},
            {"internalType": "uint256", "name": "maxPoolAmountIn", "type": "uint256"},
        ],
        "name": "exitswapExternAmountOut",
        "outputs": [
            {"internalType": "uint256", "name": "poolAmountIn", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "tokenOut", "type": "address"},
            {"internalType": "uint256", "name": "poolAmountIn", "type": "uint256"},
            {"internalType": "uint256", "name": "minAmountOut", "type": "uint256"},
        ],
        "name": "exitswapPoolAmountIn",
        "outputs": [
            {"internalType": "uint256", "name": "tokenAmountOut", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [],
        "name": "finalize",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "getBalance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getColor",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getController",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getCurrentTokens",
        "outputs": [
            {"internalType": "address[]", "name": "tokens", "type": "address[]"}
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "getDenormalizedWeight",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getFinalTokens",
        "outputs": [
            {"internalType": "address[]", "name": "tokens", "type": "address[]"}
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "getNormalizedWeight",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getNumTokens",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "address", "name": "tokenIn", "type": "address"},
            {"internalType": "address", "name": "tokenOut", "type": "address"},
        ],
        "name": "getSpotPrice",
        "outputs": [
            {"internalType": "uint256", "name": "spotPrice", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "address", "name": "tokenIn", "type": "address"},
            {"internalType": "address", "name": "tokenOut", "type": "address"},
        ],
        "name": "getSpotPriceSansFee",
        "outputs": [
            {"internalType": "uint256", "name": "spotPrice", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getSwapFee",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "getTotalDenormalizedWeight",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "gulp",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "amt", "type": "uint256"},
        ],
        "name": "increaseApproval",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "t", "type": "address"}],
        "name": "isBound",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "isFinalized",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "isPublicSwap",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "uint256", "name": "poolAmountOut", "type": "uint256"},
            {"internalType": "uint256[]", "name": "maxAmountsIn", "type": "uint256[]"},
        ],
        "name": "joinPool",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "tokenIn", "type": "address"},
            {"internalType": "uint256", "name": "tokenAmountIn", "type": "uint256"},
            {"internalType": "uint256", "name": "minPoolAmountOut", "type": "uint256"},
        ],
        "name": "joinswapExternAmountIn",
        "outputs": [
            {"internalType": "uint256", "name": "poolAmountOut", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "tokenIn", "type": "address"},
            {"internalType": "uint256", "name": "poolAmountOut", "type": "uint256"},
            {"internalType": "uint256", "name": "maxAmountIn", "type": "uint256"},
        ],
        "name": "joinswapPoolAmountOut",
        "outputs": [
            {"internalType": "uint256", "name": "tokenAmountIn", "type": "uint256"}
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "balance", "type": "uint256"},
            {"internalType": "uint256", "name": "denorm", "type": "uint256"},
        ],
        "name": "rebind",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "address", "name": "manager", "type": "address"}],
        "name": "setController",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "bool", "name": "public_", "type": "bool"}],
        "name": "setPublicSwap",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "uint256", "name": "swapFee", "type": "uint256"}],
        "name": "setSwapFee",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "tokenIn", "type": "address"},
            {"internalType": "uint256", "name": "tokenAmountIn", "type": "uint256"},
            {"internalType": "address", "name": "tokenOut", "type": "address"},
            {"internalType": "uint256", "name": "minAmountOut", "type": "uint256"},
            {"internalType": "uint256", "name": "maxPrice", "type": "uint256"},
        ],
        "name": "swapExactAmountIn",
        "outputs": [
            {"internalType": "uint256", "name": "tokenAmountOut", "type": "uint256"},
            {"internalType": "uint256", "name": "spotPriceAfter", "type": "uint256"},
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "tokenIn", "type": "address"},
            {"internalType": "uint256", "name": "maxAmountIn", "type": "uint256"},
            {"internalType": "address", "name": "tokenOut", "type": "address"},
            {"internalType": "uint256", "name": "tokenAmountOut", "type": "uint256"},
            {"internalType": "uint256", "name": "maxPrice", "type": "uint256"},
        ],
        "name": "swapExactAmountOut",
        "outputs": [
            {"internalType": "uint256", "name": "tokenAmountIn", "type": "uint256"},
            {"internalType": "uint256", "name": "spotPriceAfter", "type": "uint256"},
        ],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "amt", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "src", "type": "address"},
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "amt", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "unbind",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
]



from eth_account import Account
from packaging.version import Version

from ...eth.constants import NULL_ADDRESS
from .. import Safe
from .test_safe import TestSafe as BaseTestSafe


class TestSafeV100(BaseTestSafe):
    @property
    def safe_contract(self):
        """
        :return: Last Safe Contract available
        """
        return self.safe_contract_V1_0_0

    def deploy_test_safe(self, *args, **kwargs):
        if "fallback_handler" in kwargs:
            kwargs.pop("fallback_handler")  # It's not supported
        return self.deploy_test_safe_v1_0_0(*args, **kwargs)

    def test_create(self):
        # Create test is creating new version of the Safe
        pass

    def test_send_multisig_tx(self):
        # TODO Check this test
        pass

    def test_retrieve_fallback_handler(self):
        # Fallback handler must be empty in Safes < v1.1.0
        safe = self.deploy_test_safe()
        self.assertEqual(safe.retrieve_fallback_handler(), NULL_ADDRESS)

    def test_retrieve_guard(self):
        # Test guard must be empty in Safes < v1.2.0
        safe = self.deploy_test_safe()
        self.assertEqual(safe.retrieve_guard(), NULL_ADDRESS)
        self.assertLess(Version(safe.retrieve_version()), Version("1.3.0"))

    def test_retrieve_modules(self):
        safe = self.deploy_test_safe(owners=[self.ethereum_test_account.address])
        safe_contract = safe.contract
        module_address = Account.create().address
        self.assertEqual(safe.retrieve_modules(), [])

        tx = safe_contract.functions.enableModule(module_address).build_transaction(
            {"from": self.ethereum_test_account.address, "gas": 0, "gasPrice": 0}
        )
        safe_tx = safe.build_multisig_tx(safe.address, 0, tx["data"])
        safe_tx.sign(self.ethereum_test_account.key)
        safe_tx.execute(
            tx_sender_private_key=self.ethereum_test_account.key,
            tx_gas_price=self.gas_price,
        )
        self.assertEqual(safe.retrieve_modules(), [module_address])
        self.assertEqual(safe.retrieve_all_info().modules, [module_address])

    def test_retrieve_modules_unitialized_safe(self):
        """
        Unitialized Safes from V1.4.1 will revert
        """

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            self.ethereum_test_account,
            self.safe_contract.address,
            initializer=b"",
        )
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        self.assertEqual(safe.retrieve_modules(), [])



"""
Account abstraction utils
"""
# flake8: noqa F401
from .bundler_client import BundlerClient
from .exceptions import (
    BundlerClientConnectionException,
    BundlerClientException,
    BundlerClientResponseException,
)
from .user_operation import UserOperation, UserOperationMetadata, UserOperationV07
from .user_operation_receipt import UserOperationReceipt



from django.test import TestCase

from eth_account import Account

from gnosis.eth.constants import NULL_ADDRESS
from gnosis.eth.contracts import get_compatibility_fallback_handler_contract
from gnosis.eth.utils import fast_keccak_text

from ..signatures import get_signing_address
from .safe_test_case import SafeTestCaseMixin


class TestSafeSignature(SafeTestCaseMixin, TestCase):
    EIP1271_MAGIC_VALUE = "20c13b0b"
    EIP1271_MAGIC_VALUE_UPDATED = "1626ba7e"

    def test_get_signing_address(self):
        account = Account.create()
        # Random hash
        random_hash = fast_keccak_text("tanxugueiras")
        signature = account.signHash(random_hash)
        self.assertEqual(
            get_signing_address(random_hash, signature.v, signature.r, signature.s),
            account.address,
        )

        # Invalid signature will return the `NULL_ADDRESS`
        self.assertEqual(
            get_signing_address(random_hash, signature.v - 8, signature.r, signature.s),
            NULL_ADDRESS,
        )

    def test_eip1271_signing(self):
        owner = Account.create()
        message = "luar_na_lubre"
        safe = self.deploy_test_safe(threshold=1, owners=[owner.address])

        self.assertEqual(
            safe.contract.functions.domainSeparator().call(), safe.domain_separator
        )

        compatibility_contract = get_compatibility_fallback_handler_contract(
            self.w3, safe.address
        )
        safe_message_hash = safe.get_message_hash(message)
        self.assertEqual(
            compatibility_contract.functions.getMessageHash(message.encode()).call(),
            safe_message_hash,
        )

        # Use deprecated isValidSignature method (receives bytes)
        signature = owner.signHash(safe_message_hash)
        is_valid_bytes_fn = compatibility_contract.get_function_by_signature(
            "isValidSignature(bytes,bytes)"
        )
        self.assertEqual(
            is_valid_bytes_fn(message.encode(), signature.signature).call(),
            bytes.fromhex(self.EIP1271_MAGIC_VALUE),
        )

        # Use new isValidSignature method (receives bytes32 == hash of the message)
        # Message needs to be hashed first
        message_hash = fast_keccak_text(message)
        safe_message_hash = safe.get_message_hash(message_hash)
        self.assertEqual(
            compatibility_contract.functions.getMessageHash(message_hash).call(),
            safe_message_hash,
        )

        signature = owner.signHash(safe_message_hash)
        is_valid_bytes_fn = compatibility_contract.get_function_by_signature(
            "isValidSignature(bytes32,bytes)"
        )
        self.assertEqual(
            is_valid_bytes_fn(message_hash, signature.signature).call(),
            bytes.fromhex(self.EIP1271_MAGIC_VALUE_UPDATED),
        )

    def test_eip1271_signing_v1_1_1_contract(self):
        owner = Account.create()
        message = "luar_na_lubre"
        safe = self.deploy_test_safe_v1_1_1(threshold=1, owners=[owner.address])

        self.assertEqual(
            safe.contract.functions.domainSeparator().call(), safe.domain_separator
        )

        contract = safe.contract
        safe_message_hash = safe.get_message_hash(message)
        self.assertEqual(
            contract.functions.getMessageHash(message.encode()).call(),
            safe_message_hash,
        )

        # Use isValidSignature method (receives bytes)
        signature = owner.signHash(safe_message_hash)

        self.assertEqual(
            contract.functions.isValidSignature(
                message.encode(), signature.signature
            ).call(),
            bytes.fromhex(self.EIP1271_MAGIC_VALUE),
        )



import logging
import secrets

from django.test import TestCase

from eth_account import Account

from gnosis.eth import EthereumClient
from gnosis.eth.contracts import (
    get_proxy_1_0_0_deployed_bytecode,
    get_proxy_1_1_1_deployed_bytecode,
    get_proxy_1_3_0_deployed_bytecode,
)
from gnosis.eth.exceptions import ContractAlreadyDeployed
from gnosis.eth.tests.utils import just_test_if_mainnet_node
from gnosis.eth.utils import compare_byte_code, fast_is_checksum_address
from gnosis.safe import Safe
from gnosis.safe.proxy_factory import (
    ProxyFactory,
    ProxyFactoryV100,
    ProxyFactoryV111,
    ProxyFactoryV130,
)
from gnosis.safe.tests.safe_test_case import SafeTestCaseMixin
from gnosis.safe.tests.utils import generate_salt_nonce

logger = logging.getLogger(__name__)


class TestProxyFactory(SafeTestCaseMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.proxy_factory = ProxyFactory(
            cls.proxy_factory_contract.address, cls.ethereum_client
        )

    def test_check_proxy_code(self):
        proxy_contract_address = self.deploy_test_safe().address
        self.assertTrue(self.proxy_factory.check_proxy_code(proxy_contract_address))

        self.assertFalse(
            self.proxy_factory.check_proxy_code(self.safe_contract.address)
        )

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            self.ethereum_test_account, self.safe_contract.address
        )
        self.assertTrue(
            self.proxy_factory.check_proxy_code(ethereum_tx_sent.contract_address)
        )
        # Test every version
        versions = [
            ("1.0.0", ProxyFactoryV100, get_proxy_1_0_0_deployed_bytecode),
            ("1.1.1", ProxyFactoryV111, get_proxy_1_1_1_deployed_bytecode),
            ("1.3.0", ProxyFactoryV130, get_proxy_1_3_0_deployed_bytecode),
        ]
        for version, ProxyFactoryVersion, get_proxy_deployed_bytecode_fn in versions:
            with self.subTest(version=version):
                try:
                    deployed_proxy_tx = ProxyFactoryVersion.deploy_contract(
                        self.ethereum_client, self.ethereum_test_account
                    )
                    contract_address = deployed_proxy_tx.contract_address
                except ContractAlreadyDeployed as e:
                    contract_address = e.address

                proxy_factory = ProxyFactory(
                    contract_address,
                    self.ethereum_client,
                    version=version,
                )
                deployed_proxy_contract_tx = (
                    proxy_factory.deploy_proxy_contract_with_nonce(
                        self.ethereum_test_account, self.safe_contract.address
                    )
                )
                self.assertTrue(
                    proxy_factory.check_proxy_code(
                        deployed_proxy_contract_tx.contract_address
                    )
                )
                self.assertTrue(
                    compare_byte_code(
                        get_proxy_deployed_bytecode_fn(),
                        self.w3.eth.get_code(
                            deployed_proxy_contract_tx.contract_address
                        ),
                    )
                )

    def test_check_proxy_code_mainnet(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        last_proxy_factory_address = "0xa6B71E26C5e0845f74c812102Ca7114b6a896AB2"
        proxy_factory = ProxyFactory(last_proxy_factory_address, ethereum_client)

        # Random Safes deployed by the proxy_factory
        safes = [
            "0x7f2722741F997c63133e656a70aE5Ae0614aD7f5",  # v1.0.0
            "0x655A9e6b044d6B62F393f9990ec3eA877e966e18",  # v1.1.1
            # '',  # v1.2.0
            "0xDaB5dc22350f9a6Aff03Cf3D9341aAD0ba42d2a6",  # v1.3.0
        ]

        for safe in safes:
            with self.subTest(safe=safe):
                self.assertTrue(proxy_factory.check_proxy_code(safe))

    def test_calculate_proxy_address(self):
        salt_nonce = secrets.randbits(256)
        address = self.proxy_factory.calculate_proxy_address(
            self.safe_contract_V1_4_1.address, b"", salt_nonce
        )
        self.assertTrue(fast_is_checksum_address(address))
        # Same call with same parameters should return the same address
        same_address = self.proxy_factory.calculate_proxy_address(
            self.safe_contract_V1_4_1.address, b"", salt_nonce
        )
        self.assertEqual(address, same_address)
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            self.ethereum_test_account,
            self.safe_contract_V1_4_1.address,
            initializer=b"",
            salt_nonce=salt_nonce,
        )
        self.assertEqual(ethereum_tx_sent.contract_address, address)

        # Calculating the proxy address after deployment should return the same address
        address_after_deploying = self.proxy_factory.calculate_proxy_address(
            self.safe_contract_V1_4_1.address, b"", salt_nonce
        )
        self.assertEqual(ethereum_tx_sent.contract_address, address_after_deploying)

        chain_specific_address = self.proxy_factory.calculate_proxy_address(
            self.safe_contract_V1_4_1.address, b"", salt_nonce, chain_specific=True
        )
        self.assertTrue(fast_is_checksum_address(chain_specific_address))
        self.assertNotEqual(address, chain_specific_address)
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            self.ethereum_test_account,
            self.safe_contract_V1_4_1.address,
            initializer=b"",
            salt_nonce=salt_nonce,
            chain_specific=True,
        )
        self.assertEqual(ethereum_tx_sent.contract_address, chain_specific_address)

    def test_deploy_proxy_contract_with_nonce(self):
        salt_nonce = generate_salt_nonce()
        owners = [Account.create().address for _ in range(2)]
        threshold = 2
        payment_token = None
        safe_create2_tx = Safe.build_safe_create2_tx(
            self.ethereum_client,
            self.safe_contract.address,
            self.proxy_factory_contract.address,
            salt_nonce,
            owners,
            threshold,
            self.gas_price,
            payment_token,
        )
        # Send ether for safe deploying costs
        self.send_tx(
            {"to": safe_create2_tx.safe_address, "value": safe_create2_tx.payment},
            self.ethereum_test_account,
        )

        proxy_factory = ProxyFactory(
            self.proxy_factory_contract.address, self.ethereum_client
        )
        ethereum_tx_sent = proxy_factory.deploy_proxy_contract_with_nonce(
            self.ethereum_test_account,
            safe_create2_tx.master_copy_address,
            safe_create2_tx.safe_setup_data,
            salt_nonce,
            safe_create2_tx.gas,
            gas_price=self.gas_price,
        )
        receipt = self.ethereum_client.get_transaction_receipt(
            ethereum_tx_sent.tx_hash, timeout=20
        )
        self.assertEqual(receipt["status"], 1)
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        self.assertEqual(
            ethereum_tx_sent.contract_address, safe_create2_tx.safe_address
        )
        self.assertEqual(set(safe.retrieve_owners()), set(owners))
        self.assertEqual(
            safe.retrieve_master_copy_address(), safe_create2_tx.master_copy_address
        )

    def test_get_proxy_runtime_code(self):
        with self.assertRaises(NotImplementedError):
            self.proxy_factory.get_proxy_runtime_code()



import dataclasses
import math
import os
from abc import ABCMeta, abstractmethod
from functools import cached_property
from logging import getLogger
from typing import Callable, Dict, List, Optional, Union

import eth_abi
from eth_abi.exceptions import DecodingError
from eth_abi.packed import encode_packed
from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress, Hash32, HexStr
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.exceptions import Web3Exception
from web3.types import BlockIdentifier, TxParams

from gnosis.eth import EthereumClient, EthereumTxSent
from gnosis.eth.constants import GAS_CALL_DATA_BYTE, NULL_ADDRESS, SENTINEL_ADDRESS
from gnosis.eth.contracts import (
    ContractBase,
    get_compatibility_fallback_handler_contract,
    get_safe_contract,
    get_safe_V0_0_1_contract,
    get_safe_V1_0_0_contract,
    get_safe_V1_1_1_contract,
    get_safe_V1_3_0_contract,
    get_safe_V1_4_1_contract,
    get_simulate_tx_accessor_V1_4_1_contract,
)
from gnosis.eth.utils import (
    fast_bytes_to_checksum_address,
    fast_is_checksum_address,
    fast_keccak,
    get_empty_tx_params,
)

from ..eth.typing import EthereumData
from .addresses import SAFE_SIMULATE_TX_ACCESSOR_ADDRESS
from .enums import SafeOperationEnum
from .exceptions import CannotEstimateGas, CannotRetrieveSafeInfoException
from .safe_creator import SafeCreator
from .safe_tx import SafeTx

logger = getLogger(__name__)


@dataclasses.dataclass
class SafeInfo:
    address: ChecksumAddress
    fallback_handler: ChecksumAddress
    guard: ChecksumAddress
    master_copy: ChecksumAddress
    modules: List[ChecksumAddress]
    nonce: int
    owners: List[ChecksumAddress]
    threshold: int
    version: str


class Safe(SafeCreator, ContractBase, metaclass=ABCMeta):
    """
    Collection of methods and utilies to handle a Safe
    """

    # keccak256("fallback_manager.handler.address")
    FALLBACK_HANDLER_STORAGE_SLOT = (
        0x6C9A6C4A39284E37ED1CF53D337577D14212A4870FB976A4366C693B939918D5
    )
    # keccak256("guard_manager.guard.address")
    GUARD_STORAGE_SLOT = (
        0x4A204F620C8C5CCDCA3FD54D003BADD85BA500436A431F0CBDA4F558C93C34C8
    )

    # keccak256("SafeMessage(bytes message)");
    SAFE_MESSAGE_TYPEHASH = bytes.fromhex(
        "60b3cbf8b4a223d68d641b3b6ddf9a298e7f33710cf3d3a9d1146b5a6150fbca"
    )

    def __new__(
        cls, address: ChecksumAddress, ethereum_client: EthereumClient, *args, **kwargs
    ) -> "Safe":
        """
        Hacky factory for Safe

        :param address:
        :param ethereum_client:
        :param kwargs:
        """
        assert fast_is_checksum_address(address), "%s is not a valid address" % address
        if cls is not Safe:
            return super().__new__(cls, *args, **kwargs)

        versions: Dict[str, Safe] = {
            "0.0.1": SafeV001,
            "1.0.0": SafeV100,
            "1.1.1": SafeV111,
            "1.2.0": SafeV120,
            "1.3.0": SafeV130,
            "1.4.1": SafeV141,
        }
        default_version = SafeV141

        version: Optional[str]
        try:
            contract = get_safe_contract(ethereum_client.w3, address=address)
            version = contract.functions.VERSION().call(block_identifier="latest")
        except (Web3Exception, ValueError):
            version = None  # Cannot detect the version

        instance_class = versions.get(version, default_version)
        instance = super().__new__(instance_class)
        return instance

    def __init__(
        self,
        address: ChecksumAddress,
        ethereum_client: EthereumClient,
        simulate_tx_accessor_address: Optional[ChecksumAddress] = None,
    ):
        self._simulate_tx_accessor_address = simulate_tx_accessor_address
        super().__init__(address, ethereum_client)

    def __str__(self):
        return f"Safe={self.address}"

    @abstractmethod
    def get_version(self) -> str:
        """
        :return: String with Safe Master Copy semantic version, must match `retrieve_version()`
        """
        raise NotImplementedError

    @cached_property
    def chain_id(self) -> int:
        return self.ethereum_client.get_chain_id()

    @property
    def simulate_tx_accessor_address(self) -> ChecksumAddress:
        if self._simulate_tx_accessor_address:
            return self._simulate_tx_accessor_address
        return os.environ.get(
            "SAFE_SIMULATE_TX_ACCESSOR_ADDRESS", SAFE_SIMULATE_TX_ACCESSOR_ADDRESS
        )

    @simulate_tx_accessor_address.setter
    def simulate_tx_accessor_address(self, value: ChecksumAddress):
        self._simulate_tx_accessor_address = value

    def retrieve_version(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> str:
        return self.contract.functions.VERSION().call(block_identifier=block_identifier)

    @cached_property
    def domain_separator(self) -> Optional[bytes]:
        """
        :return: EIP721 DomainSeparator for the Safe. Returns `None` if not supported (for Safes < 1.0.0)
        """
        try:
            return self.retrieve_domain_separator()
        except (Web3Exception, DecodingError, ValueError):
            logger.warning("Safe %s does not support domainSeparator", self.address)
            return None

    @classmethod
    def deploy_contract(
        cls,
        ethereum_client: EthereumClient,
        deployer_account: LocalAccount,
    ) -> EthereumTxSent:
        """
        Deploy master contract. Takes deployer_account (if unlocked in the node) or the deployer private key
        Safe with version > v1.1.1 doesn't need to be initialized as it already has a constructor

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: ``EthereumTxSent`` with the deployed contract address
        """
        contract_fn = cls.get_contract_fn(cls)
        safe_contract = contract_fn(ethereum_client.w3)
        constructor_data = safe_contract.constructor().build_transaction(
            get_empty_tx_params()
        )["data"]
        ethereum_tx_sent = ethereum_client.deploy_and_initialize_contract(
            deployer_account, constructor_data
        )
        deployed_version = (
            contract_fn(ethereum_client.w3, ethereum_tx_sent.contract_address)
            .functions.VERSION()
            .call()
        )
        assert deployed_version == cls.get_version(
            cls
        ), f"Deployed version {deployed_version} is not matching expected {cls.get_version(cls)} version"

        logger.info(
            "Deployed and initialized Safe Master Contract version=%s on address %s by %s",
            deployed_version,
            ethereum_tx_sent.contract_address,
            deployer_account.address,
        )
        return ethereum_tx_sent

    def check_funds_for_tx_gas(
        self, safe_tx_gas: int, base_gas: int, gas_price: int, gas_token: str
    ) -> bool:
        """
        Check safe has enough funds to pay for a tx

        :param safe_tx_gas: Safe tx gas
        :param base_gas: Data gas
        :param gas_price: Gas Price
        :param gas_token: Gas Token, to use token instead of ether for the gas
        :return: `True` if enough funds, `False` otherwise
        """
        if gas_token == NULL_ADDRESS:
            balance = self.ethereum_client.get_balance(self.address)
        else:
            balance = self.ethereum_client.erc20.get_balance(self.address, gas_token)
        return balance >= (safe_tx_gas + base_gas) * gas_price

    def estimate_tx_base_gas(
        self,
        to: ChecksumAddress,
        value: int,
        data: bytes,
        operation: int,
        gas_token: ChecksumAddress,
        estimated_tx_gas: int,
    ) -> int:
        """
        Calculate gas costs that are independent of the transaction execution(e.g. base transaction fee,
        signature check, payment of the refund...)

        :param to:
        :param value:
        :param data:
        :param operation:
        :param gas_token:
        :param estimated_tx_gas: gas calculated with `estimate_tx_gas`
        :return:
        """
        data = data or b""
        safe_contract = self.contract
        threshold = self.retrieve_threshold()
        nonce = self.retrieve_nonce()

        # Every byte == 0 -> 4  Gas
        # Every byte != 0 -> 16 Gas (68 before Istanbul)
        # numbers < 256 (0x00(31*2)..ff) are 192 -> 31 * 4 + 1 * GAS_CALL_DATA_BYTE
        # numbers < 65535 (0x(30*2)..ffff) are 256 -> 30 * 4 + 2 * GAS_CALL_DATA_BYTE

        # Calculate gas for signatures
        # (array count (3 -> r, s, v) + ecrecover costs) * signature count
        # ecrecover for ecdsa ~= 4K gas, we use 6K
        ecrecover_gas = 6000
        signature_gas = threshold * (
            1 * GAS_CALL_DATA_BYTE + 2 * 32 * GAS_CALL_DATA_BYTE + ecrecover_gas
        )

        safe_tx_gas = estimated_tx_gas
        base_gas = 0
        gas_price = 1
        gas_token = gas_token or NULL_ADDRESS
        signatures = b""
        refund_receiver = NULL_ADDRESS
        data = HexBytes(
            safe_contract.functions.execTransaction(
                to,
                value,
                data,
                operation,
                safe_tx_gas,
                base_gas,
                gas_price,
                gas_token,
                refund_receiver,
                signatures,
            ).build_transaction(get_empty_tx_params())["data"]
        )

        # If nonce == 0, nonce storage has to be initialized
        if nonce == 0:
            nonce_gas = 20000
        else:
            nonce_gas = 5000

        # Keccak costs for the hash of the safe tx
        hash_generation_gas = 1500

        base_gas = (
            signature_gas
            + self.ethereum_client.estimate_data_gas(data)
            + nonce_gas
            + hash_generation_gas
        )

        # Add additional gas costs
        if base_gas > 65536:
            base_gas += 64
        else:
            base_gas += 128

        base_gas += 32000  # Base tx costs, transfer costs...
        return base_gas

    def estimate_tx_gas_with_safe(
        self,
        to: ChecksumAddress,
        value: int,
        data: bytes,
        operation: int,
        gas_limit: Optional[int] = None,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> int:
        """
        Estimate tx gas using safe `requiredTxGas` method

        :return: int: Estimated gas
        :raises: CannotEstimateGas: If gas cannot be estimated
        :raises: ValueError: Cannot decode received data
        """

        safe_address = self.address
        data = data or b""

        def parse_revert_data(result: bytes) -> int:
            # 4 bytes - error method id
            # 32 bytes - position
            # 32 bytes - length
            # Last 32 bytes - value of revert (if everything went right)
            gas_estimation_offset = 4 + 32 + 32
            gas_estimation = result[gas_estimation_offset:]

            # Estimated gas must be 32 bytes
            if len(gas_estimation) != 32:
                gas_limit_text = (
                    f"with gas limit={gas_limit} "
                    if gas_limit is not None
                    else "without gas limit set "
                )
                logger.warning(
                    "Safe=%s Problem estimating gas, returned value %sis %s for tx=%s",
                    safe_address,
                    gas_limit_text,
                    result.hex(),
                    tx,
                )
                raise CannotEstimateGas("Received %s for tx=%s" % (result.hex(), tx))

            return int(gas_estimation.hex(), 16)

        tx = self.contract.functions.requiredTxGas(
            to, value, data, operation
        ).build_transaction(
            {
                "from": safe_address,
                "gas": 0,  # Don't call estimate
                "gasPrice": 0,  # Don't get gas price
            }
        )

        tx_params = {
            "from": safe_address,
            "to": safe_address,
            "data": tx["data"],
        }

        if gas_limit:
            tx_params["gas"] = hex(gas_limit)

        query = {
            "jsonrpc": "2.0",
            "method": "eth_call",
            "params": [tx_params, block_identifier],
            "id": 1,
        }

        response = self.ethereum_client.http_session.post(
            self.ethereum_client.ethereum_node_url, json=query, timeout=30
        )
        if response.ok:
            response_data = response.json()
            error_data: Optional[str] = None
            if "error" in response_data and "data" in response_data["error"]:
                error_data = response_data["error"]["data"]
            elif "result" in response_data:  # Ganache-cli
                error_data = response_data["result"]

            if error_data:
                if "0x" in error_data:
                    return parse_revert_data(
                        HexBytes(error_data[error_data.find("0x") :])
                    )

        raise CannotEstimateGas(
            f"Received {response.status_code} - {response.content} from ethereum node"
        )

    def estimate_tx_gas_with_web3(
        self, to: ChecksumAddress, value: int, data: EthereumData
    ) -> int:
        """
        :param to:
        :param value:
        :param data:
        :return: Estimation using web3 `estimate_gas`
        """
        try:
            return self.ethereum_client.estimate_gas(
                to, from_=self.address, value=value, data=data
            )
        except (Web3Exception, ValueError) as exc:
            raise CannotEstimateGas(
                f"Cannot estimate gas with `eth_estimateGas`: {exc}"
            ) from exc

    def estimate_tx_gas_by_trying(
        self, to: ChecksumAddress, value: int, data: Union[bytes, str], operation: int
    ):
        """
        Try to get an estimation with Safe's `requiredTxGas`. If estimation is successful, try to set a gas limit and
        estimate again. If gas estimation is ok, same gas estimation should be returned, if it's less than required
        estimation will not be completed, so estimation was not accurate and gas limit needs to be increased.

        :param to:
        :param value:
        :param data:
        :param operation:
        :return: Estimated gas calling `requiredTxGas` setting a gas limit and checking if `eth_call` is successful
        :raises: CannotEstimateGas
        """
        if not data:
            data = b""
        elif isinstance(data, str):
            data = HexBytes(data)

        gas_estimated = self.estimate_tx_gas_with_safe(to, value, data, operation)
        block_gas_limit: Optional[int] = None
        base_gas: Optional[int] = self.ethereum_client.estimate_data_gas(data)

        for i in range(
            1, 30
        ):  # Make sure tx can be executed, fixing for example 63/64th problem
            try:
                self.estimate_tx_gas_with_safe(
                    to,
                    value,
                    data,
                    operation,
                    gas_limit=gas_estimated + base_gas + 32000,
                )
                return gas_estimated
            except CannotEstimateGas:
                logger.warning(
                    "Safe=%s - Found 63/64 problem gas-estimated=%d to=%s data=%s",
                    self.address,
                    gas_estimated,
                    to,
                    data.hex(),
                )
                block_gas_limit = (
                    block_gas_limit
                    or self.w3.eth.get_block("latest", full_transactions=False)[
                        "gasLimit"
                    ]
                )
                gas_estimated = math.floor((1 + i * 0.03) * gas_estimated)
                if gas_estimated >= block_gas_limit:
                    return block_gas_limit
        return gas_estimated

    def estimate_tx_gas(
        self, to: ChecksumAddress, value: int, data: bytes, operation: int
    ) -> int:
        """
        Estimate tx gas. Use `requiredTxGas` on the Safe contract and fallbacks to `eth_estimateGas` if that method
        fails. Note: `eth_estimateGas` cannot estimate delegate calls

        :param to:
        :param value:
        :param data:
        :param operation:
        :return: Estimated gas for Safe inner tx
        :raises: CannotEstimateGas
        """
        # Costs to route through the proxy and nested calls
        PROXY_GAS = 1000
        # https://github.com/ethereum/solidity/blob/dfe3193c7382c80f1814247a162663a97c3f5e67/libsolidity/codegen/ExpressionCompiler.cpp#L1764
        # This was `false` before solc 0.4.21 -> `m_context.evmVersion().canOverchargeGasForCall()`
        # So gas needed by caller will be around 35k
        OLD_CALL_GAS = 35000
        # Web3 `estimate_gas` estimates less gas
        WEB3_ESTIMATION_OFFSET = 23000
        ADDITIONAL_GAS = PROXY_GAS + OLD_CALL_GAS

        try:
            return (
                self.estimate_tx_gas_by_trying(to, value, data, operation)
                + ADDITIONAL_GAS
            )
        except CannotEstimateGas:
            return (
                self.estimate_tx_gas_with_web3(to, value, data)
                + ADDITIONAL_GAS
                + WEB3_ESTIMATION_OFFSET
            )

    def get_message_hash(self, message: Union[str, Hash32]) -> Hash32:
        """
        Return hash of a message that can be signed by owners.

        :param message: Message that should be hashed. A ``Hash32`` must be provided for EIP191 or EIP712 messages
        :return: Message hash
        """

        if isinstance(message, str):
            message = message.encode()
        message_hash = fast_keccak(message)

        safe_message_hash = fast_keccak(
            eth_abi.encode(
                ["bytes32", "bytes32"], [self.SAFE_MESSAGE_TYPEHASH, message_hash]
            )
        )
        return fast_keccak(
            encode_packed(
                ["bytes1", "bytes1", "bytes32", "bytes32"],
                [
                    bytes.fromhex("19"),
                    bytes.fromhex("01"),
                    self.domain_separator,
                    safe_message_hash,
                ],
            )
        )

    def retrieve_all_info(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> SafeInfo:
        """
        Get all Safe info in the same batch call.

        :param block_identifier:
        :return:
        :raises: CannotRetrieveSafeInfoException
        """

        # FIXME for not initialized Safes `getModules` get into an infinite loop on the RPC
        try:
            contract = self.contract
            master_copy = self.retrieve_master_copy_address()
            if master_copy == NULL_ADDRESS:
                raise CannotRetrieveSafeInfoException(self.address)

            fallback_handler = self.retrieve_fallback_handler()
            guard = self.retrieve_guard()  # Guard was implemented in v1.1.1

            # From v1.1.1:
            # - `getModulesPaginated` is available
            # - `getModules` returns only 10 modules
            modules_fn = (
                contract.functions.getModulesPaginated(SENTINEL_ADDRESS, 20)
                if hasattr(contract.functions, "getModulesPaginated")
                else contract.functions.getModules()
            )

            results = self.ethereum_client.batch_call(
                [
                    modules_fn,
                    contract.functions.nonce(),
                    contract.functions.getOwners(),
                    contract.functions.getThreshold(),
                    contract.functions.VERSION(),
                ],
                from_address=self.address,
                block_identifier=block_identifier,
                raise_exception=False,
            )
            modules_response, nonce, owners, threshold, version = results
            if (
                modules_response
                and len(modules_response) == 2
                and isinstance(modules_response[0], (tuple, list))
            ):
                # Must be a Tuple[List[ChecksumAddress], ChecksumAddress]
                # >= v1.1.1
                modules, next_module = modules_response
                if modules and next_module != SENTINEL_ADDRESS:
                    # Still more elements in the list
                    modules = self.retrieve_modules()
            else:
                # < v1.1.1
                modules = modules_response

            return SafeInfo(
                self.address,
                fallback_handler,
                guard,
                master_copy,
                modules if modules else [],
                nonce,
                owners,
                threshold,
                version,
            )
        except (Web3Exception, ValueError) as e:
            raise CannotRetrieveSafeInfoException(self.address) from e

    def retrieve_domain_separator(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> str:
        return self.contract.functions.domainSeparator().call(
            block_identifier=block_identifier
        )

    def retrieve_code(self) -> HexBytes:
        return self.w3.eth.get_code(self.address)

    def retrieve_fallback_handler(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> ChecksumAddress:
        address = self.ethereum_client.w3.eth.get_storage_at(
            self.address,
            self.FALLBACK_HANDLER_STORAGE_SLOT,
            block_identifier=block_identifier,
        )[-20:].rjust(20, b"\0")
        if len(address) == 20:
            return fast_bytes_to_checksum_address(address)
        else:
            return NULL_ADDRESS

    def retrieve_guard(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> ChecksumAddress:
        address = self.ethereum_client.w3.eth.get_storage_at(
            self.address, self.GUARD_STORAGE_SLOT, block_identifier=block_identifier
        )[-20:].rjust(20, b"\0")
        if len(address) == 20:
            return fast_bytes_to_checksum_address(address)
        else:
            return NULL_ADDRESS

    def retrieve_master_copy_address(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> ChecksumAddress:
        address = self.w3.eth.get_storage_at(
            self.address, "0x00", block_identifier=block_identifier
        )[-20:].rjust(20, b"\0")
        return fast_bytes_to_checksum_address(address)

    def retrieve_modules(
        self,
        pagination: Optional[int] = 50,
        max_modules_to_retrieve: Optional[int] = 500,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> List[ChecksumAddress]:
        """
        Get modules enabled on the Safe
        From v1.1.1:
          - ``getModulesPaginated`` is available
          - ``getModules`` returns only 10 modules

        :param pagination: Number of modules to get per request
        :param max_modules_to_retrieve: Maximum number of modules to retrieve
        :param block_identifier:
        :return: List of module addresses
        """
        if not hasattr(self.contract.functions, "getModulesPaginated"):
            # Custom code for Safes < v1.3.0
            # Safe V1_0_0 can get into an infinite loop if it's not initialized
            if self.retrieve_threshold() == 0:
                return []
            return self.contract.functions.getModules().call(
                block_identifier=block_identifier
            )

        # We need to iterate the module paginator
        contract = self.contract
        next_module = SENTINEL_ADDRESS
        all_modules: List[ChecksumAddress] = []

        for _ in range(max_modules_to_retrieve // pagination):
            # If we use a `while True` loop a custom coded Safe could get us into an infinite loop
            (modules, next_module) = contract.functions.getModulesPaginated(
                next_module, pagination
            ).call(block_identifier=block_identifier)

            # Safes with version < 1.4.0 don't include the `starter address` used as pagination in the module list
            # From 1.4.0 onwards it is included, so we check for duplicated addresses before inserting
            for module in modules + [next_module]:
                if module not in all_modules + [NULL_ADDRESS, SENTINEL_ADDRESS]:
                    all_modules.append(module)

            if not modules or next_module in (NULL_ADDRESS, SENTINEL_ADDRESS):
                # `NULL_ADDRESS` is only seen in uninitialized Safes
                break
        return all_modules

    def retrieve_is_hash_approved(
        self,
        owner: str,
        safe_hash: bytes,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> bool:
        return (
            self.contract.functions.approvedHashes(owner, safe_hash).call(
                block_identifier=block_identifier
            )
            == 1
        )

    def retrieve_is_message_signed(
        self,
        message_hash: Hash32,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> bool:
        return self.contract.functions.signedMessages(message_hash).call(
            block_identifier=block_identifier
        )

    def retrieve_is_owner(
        self, owner: str, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> bool:
        return self.contract.functions.isOwner(owner).call(
            block_identifier=block_identifier
        )

    def retrieve_nonce(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> int:
        return self.contract.functions.nonce().call(block_identifier=block_identifier)

    def retrieve_owners(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> List[str]:
        return self.contract.functions.getOwners().call(
            block_identifier=block_identifier
        )

    def retrieve_threshold(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> int:
        return self.contract.functions.getThreshold().call(
            block_identifier=block_identifier
        )

    def build_multisig_tx(
        self,
        to: ChecksumAddress,
        value: int,
        data: bytes,
        operation: int = SafeOperationEnum.CALL.value,
        safe_tx_gas: int = 0,
        base_gas: int = 0,
        gas_price: int = 0,
        gas_token: ChecksumAddress = NULL_ADDRESS,
        refund_receiver: ChecksumAddress = NULL_ADDRESS,
        signatures: bytes = b"",
        safe_nonce: Optional[int] = None,
    ) -> SafeTx:
        """
        Allows to execute a Safe transaction confirmed by required number of owners and then pays the account
        that submitted the transaction. The fees are always transfered, even if the user transaction fails

        :param to: Destination address of Safe transaction
        :param value: Ether value of Safe transaction
        :param data: Data payload of Safe transaction
        :param operation: Operation type of Safe transaction
        :param safe_tx_gas: Gas that should be used for the Safe transaction
        :param base_gas: Gas costs for that are independent of the transaction execution
            (e.g. base transaction fee, signature check, payment of the refund)
        :param gas_price: Gas price that should be used for the payment calculation
        :param gas_token: Token address (or `0x000..000` if ETH) that is used for the payment
        :param refund_receiver: Address of receiver of gas payment (or `0x000..000` if tx.origin).
        :param signatures: Packed signature data ({bytes32 r}{bytes32 s}{uint8 v})
        :param safe_nonce: Nonce of the safe (to calculate hash)
        :param safe_version: Safe version (to calculate hash)
        :return: SafeTx
        """

        if safe_nonce is None:
            safe_nonce = self.retrieve_nonce()
        return SafeTx(
            self.ethereum_client,
            self.address,
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            signatures=signatures,
            safe_nonce=safe_nonce,
            safe_version=self.get_version(),
            chain_id=self.chain_id,
        )

    def send_multisig_tx(
        self,
        to: ChecksumAddress,
        value: int,
        data: bytes,
        operation: int,
        safe_tx_gas: int,
        base_gas: int,
        gas_price: int,
        gas_token: ChecksumAddress,
        refund_receiver: ChecksumAddress,
        signatures: bytes,
        tx_sender_private_key: HexStr,
        tx_gas=None,
        tx_gas_price=None,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> EthereumTxSent:
        """
        Build and send Safe tx

        :param to:
        :param value:
        :param data:
        :param operation:
        :param safe_tx_gas:
        :param base_gas:
        :param gas_price:
        :param gas_token:
        :param refund_receiver:
        :param signatures:
        :param tx_sender_private_key:
        :param tx_gas: Gas for the external tx. If not, `(safe_tx_gas + data_gas) * 2` will be used
        :param tx_gas_price: Gas price of the external tx. If not, `gas_price` will be used
        :param block_identifier:
        :return: Tuple(tx_hash, tx)
        :raises: InvalidMultisigTx: If user tx cannot go through the Safe
        """

        safe_tx = self.build_multisig_tx(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            signatures,
        )

        tx_sender_address = Account.from_key(tx_sender_private_key).address
        safe_tx.call(
            tx_sender_address=tx_sender_address, block_identifier=block_identifier
        )

        tx_hash, tx = safe_tx.execute(
            tx_sender_private_key=tx_sender_private_key,
            tx_gas=tx_gas,
            tx_gas_price=tx_gas_price,
            block_identifier=block_identifier,
        )

        return EthereumTxSent(tx_hash, tx, None)


class SafeV001(Safe):
    def get_version(self):
        return "0.0.1"

    def get_contract_fn(self) -> Callable[[Web3, ChecksumAddress], Contract]:
        return get_safe_V0_0_1_contract

    @staticmethod
    def deploy_contract(
        ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy master contract. Takes deployer_account (if unlocked in the node) or the deployer private key

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: ``EthereumTxSent`` with the deployed contract address
        """

        safe_contract = get_safe_V0_0_1_contract(ethereum_client.w3)
        constructor_data = safe_contract.constructor().build_transaction(
            get_empty_tx_params()
        )["data"]
        initializer_data = safe_contract.functions.setup(
            # We use 2 owners that nobody controls for the master copy
            [
                "0x0000000000000000000000000000000000000002",
                "0x0000000000000000000000000000000000000003",
            ],
            2,  # Threshold. Maximum security
            NULL_ADDRESS,  # Address for optional DELEGATE CALL
            b"",  # Data for optional DELEGATE CALL
        ).build_transaction({"to": NULL_ADDRESS, "gas": 0, "gasPrice": 0})["data"]

        ethereum_tx_sent = ethereum_client.deploy_and_initialize_contract(
            deployer_account, constructor_data, HexBytes(initializer_data)
        )
        logger.info(
            "Deployed and initialized Old Safe Master Contract=%s by %s",
            ethereum_tx_sent.contract_address,
            deployer_account.address,
        )
        return ethereum_tx_sent


class SafeV100(Safe):
    def get_version(self):
        return "1.0.0"

    def get_contract_fn(self) -> Contract:
        return get_safe_V1_0_0_contract

    @staticmethod
    def deploy_contract(
        ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy master contract. Takes deployer_account (if unlocked in the node) or the deployer private key

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: ``EthereumTxSent`` with the deployed contract address
        """

        safe_contract = get_safe_V1_0_0_contract(ethereum_client.w3)
        constructor_data = safe_contract.constructor().build_transaction(
            get_empty_tx_params()
        )["data"]
        initializer_data = safe_contract.functions.setup(
            # We use 2 owners that nobody controls for the master copy
            [
                "0x0000000000000000000000000000000000000002",
                "0x0000000000000000000000000000000000000003",
            ],
            2,  # Threshold. Maximum security
            NULL_ADDRESS,  # Address for optional DELEGATE CALL
            b"",  # Data for optional DELEGATE CALL
            NULL_ADDRESS,  # Payment token
            0,  # Payment
            NULL_ADDRESS,  # Refund receiver
        ).build_transaction({"to": NULL_ADDRESS, "gas": 0, "gasPrice": 0})["data"]

        ethereum_tx_sent = ethereum_client.deploy_and_initialize_contract(
            deployer_account, constructor_data, HexBytes(initializer_data)
        )
        logger.info(
            "Deployed and initialized Safe Master Contract=%s by %s",
            ethereum_tx_sent.contract_address,
            deployer_account.address,
        )
        return ethereum_tx_sent


class SafeV111(Safe):
    def get_version(self):
        return "1.1.1"

    def get_contract_fn(self) -> Contract:
        return get_safe_V1_1_1_contract


class SafeV120(Safe):
    def get_version(self):
        return "1.2.0"

    def get_contract_fn(self) -> Contract:
        return get_safe_V1_1_1_contract


class SafeV130(Safe):
    def get_version(self):
        return "1.3.0"

    def get_contract_fn(self) -> Contract:
        return get_safe_V1_3_0_contract


class SafeV141(Safe):
    def get_version(self):
        return "1.4.1"

    def get_contract_fn(self) -> Contract:
        return get_safe_V1_4_1_contract

    def estimate_tx_gas_with_safe(
        self,
        to: ChecksumAddress,
        value: int,
        data: bytes,
        operation: int,
        gas_limit: Optional[int] = None,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> int:
        """
        Estimate tx gas. Use `SimulateTxAccesor` and `simulate` on the `CompatibilityFallHandler`

        :param to:
        :param value:
        :param data:
        :param operation:
        :param gas_limit:
        :param block_identifier:
        :return:
        """
        accessor = get_simulate_tx_accessor_V1_4_1_contract(
            self.w3, address=self.simulate_tx_accessor_address
        )
        simulator = get_compatibility_fallback_handler_contract(
            self.w3, address=self.address
        )
        simulation_data = accessor.functions.simulate(
            to, value, data, operation
        ).build_transaction(get_empty_tx_params())["data"]
        params: TxParams = {"gas": gas_limit} if gas_limit else {}
        # params = {'gas': 2_045_741}
        try:
            accessible_data = simulator.functions.simulate(
                accessor.address, simulation_data
            ).call(params)
        except ValueError as e:
            raise CannotEstimateGas(f"Reverted call using SimulateTxAccessor {e}")
        try:
            # Simulate returns (uint256 estimate, bool success, bytes memory returnData)
            (estimate, success, return_data) = eth_abi.decode(
                ["uint256", "bool", "bytes"], accessible_data
            )
            if not success:
                raise CannotEstimateGas(
                    "Cannot estimate gas using SimulateTxAccessor - Execution not successful"
                )
            return estimate
        except DecodingError as e:
            try:
                decoded_revert = eth_abi.decode(["string"], accessible_data)
            except DecodingError:
                decoded_revert = "No revert message"
            raise CannotEstimateGas(
                f"Cannot estimate gas using SimulateTxAccessor {e} - {decoded_revert}"
            )



import os
import random

from gnosis.eth.constants import SECPK1_N


def generate_salt_nonce() -> int:
    return random.getrandbits(256) - 1


def generate_valid_s() -> int:
    while True:
        s = int(os.urandom(30).hex(), 16)
        if s <= (SECPK1_N // 2):
            return s



super_token_abi = [
    {
        "inputs": [],
        "name": "getUnderlyingToken",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
]



from abc import ABC
from typing import Dict, Optional
from urllib.parse import urljoin

import requests

from gnosis.eth.ethereum_client import (
    EthereumClient,
    EthereumNetwork,
    EthereumNetworkNotSupported,
)
from gnosis.util.http import prepare_http_session


class SafeAPIException(Exception):
    pass


class SafeBaseAPI(ABC):
    URL_BY_NETWORK: Dict[EthereumNetwork, str] = {}

    def __init__(
        self,
        network: EthereumNetwork,
        ethereum_client: Optional[EthereumClient] = None,
        base_url: Optional[str] = None,
        request_timeout: int = 10,
    ):
        """
        :param network: Network for the transaction service
        :param ethereum_client:
        :param base_url: If a custom transaction service is used
        :raises: EthereumNetworkNotSupported
        """
        self.network = network
        self.ethereum_client = ethereum_client
        self.base_url = base_url or self.URL_BY_NETWORK.get(network)
        if not self.base_url:
            raise EthereumNetworkNotSupported(network)
        self.http_session = prepare_http_session(10, 100)
        self.request_timeout = request_timeout

    @classmethod
    def from_ethereum_client(cls, ethereum_client: EthereumClient) -> "SafeBaseAPI":
        ethereum_network = ethereum_client.get_network()
        return cls(ethereum_network, ethereum_client=ethereum_client)

    def _get_request(self, url: str) -> requests.Response:
        full_url = urljoin(self.base_url, url)
        return self.http_session.get(full_url, timeout=self.request_timeout)

    def _post_request(self, url: str, payload: Dict) -> requests.Response:
        full_url = urljoin(self.base_url, url)
        return self.http_session.post(
            full_url,
            json=payload,
            headers={"Content-type": "application/json"},
            timeout=self.request_timeout,
        )

    def _delete_request(self, url: str, payload: Dict) -> requests.Response:
        full_url = urljoin(self.base_url, url)
        return self.http_session.delete(
            full_url,
            json=payload,
            headers={"Content-type": "application/json"},
            timeout=self.request_timeout,
        )



from django.core.exceptions import ValidationError

from ..utils import fast_is_checksum_address


def validate_checksumed_address(address):
    if not fast_is_checksum_address(address):
        raise ValidationError(
            "%(address)s has an invalid checksum",
            params={"address": address},
        )



import requests


def prepare_http_session(
    pool_connections: int,
    pool_maxsize: int,
    retry_count: int = 0,
    pool_block: bool = False,
) -> requests.Session:
    """
    Prepare http session with custom pooling. See:
    https://urllib3.readthedocs.io/en/stable/advanced-usage.html
    https://docs.python-requests.org/en/latest/api/#requests.adapters.HTTPAdapter
    """
    session = requests.Session()
    retry_conf = (
        requests.adapters.Retry(
            total=retry_count, backoff_factor=0.3, respect_retry_after_header=False
        )
        if retry_count
        else 0
    )

    adapter = requests.adapters.HTTPAdapter(
        pool_connections=pool_connections,  # Doing all the connections to the same url
        pool_maxsize=pool_maxsize,  # Number of concurrent connections
        max_retries=retry_conf,  # Nodes are not very responsive some times
        pool_block=pool_block,
    )
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



from django.test import TestCase

from eth_account import Account

from ... import EthereumClient
from ...oracles import CannotGetPriceFromOracle, SuperfluidOracle, SushiswapOracle
from ..ethereum_test_case import EthereumTestCaseMixin
from ..test_oracles import gno_token_mainnet_address
from ..utils import just_test_if_polygon_node


class TestSuperfluidOracle(EthereumTestCaseMixin, TestCase):
    def test_get_price(self):
        polygon_node = just_test_if_polygon_node()

        self.assertFalse(SuperfluidOracle.is_available(self.ethereum_client))

        ethereum_client_polygon = EthereumClient(polygon_node)

        self.assertTrue(SuperfluidOracle.is_available(ethereum_client_polygon))

        sushi_oracle_polygon = SushiswapOracle(ethereum_client_polygon)
        superfluid_oracle_polygon = SuperfluidOracle(
            ethereum_client_polygon, sushi_oracle_polygon
        )
        uscdcx_address_polygon = "0xCAa7349CEA390F89641fe306D93591f87595dc1F"
        price = superfluid_oracle_polygon.get_price(uscdcx_address_polygon)
        self.assertGreater(price, 0.0)

        error_message = "It is not a wrapper Super Token"
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            superfluid_oracle_polygon.get_price(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            superfluid_oracle_polygon.get_price(Account.create().address)



from typing import List
from unittest import mock
from unittest.mock import MagicMock

from django.test import TestCase

from ... import EthereumNetwork
from ...clients import SourcifyClient
from ...clients.sourcify_client import SourcifyClientConfigurationProblem


class TestSourcifyClient(TestCase):
    @mock.patch.object(SourcifyClient, "is_chain_supported")
    def test_init(self, is_chain_supported_mock: MagicMock):
        is_chain_supported_mock.return_value = False
        with self.assertRaises(SourcifyClientConfigurationProblem):
            SourcifyClient(EthereumNetwork.OLYMPIC)

        is_chain_supported_mock.return_value = True
        self.assertIsInstance(SourcifyClient(), SourcifyClient)
        self.assertIsInstance(SourcifyClient(EthereumNetwork.GNOSIS), SourcifyClient)

    def test_is_chain_supported(self):
        try:
            sourcify = SourcifyClient()
        except IOError:
            self.skipTest("Cannot connect to Sourcify")

        self.assertTrue(sourcify.is_chain_supported(EthereumNetwork.MAINNET.value))
        self.assertTrue(sourcify.is_chain_supported(EthereumNetwork.GNOSIS.value))
        self.assertFalse(sourcify.is_chain_supported(2))

    @mock.patch.object(SourcifyClient, "is_chain_supported", return_value=True)
    def test_get_contract_metadata(self, is_chain_supported_mock: MagicMock):
        sourcify_client_mainnet = SourcifyClient()
        safe_contract_address = "0x41675C099F32341bf84BFc5382aF534df5C7461a"
        try:
            safe_contract_metadata_mainnet = (
                sourcify_client_mainnet.get_contract_metadata(safe_contract_address)
            )
        except IOError:
            self.skipTest("Cannot connect to Sourcify")
        self.assertEqual(safe_contract_metadata_mainnet.name, "Safe")
        self.assertIsInstance(safe_contract_metadata_mainnet.abi, List)
        self.assertTrue(safe_contract_metadata_mainnet.abi)
        self.assertFalse(safe_contract_metadata_mainnet.partial_match)
        sourcify_client_sepolia = SourcifyClient(EthereumNetwork.SEPOLIA)
        contract_metadata_sepolia = sourcify_client_sepolia.get_contract_metadata(
            safe_contract_address
        )
        self.assertEqual(safe_contract_metadata_mainnet, contract_metadata_sepolia)

        # Testing sourcify partial match token
        partial_match_contract_address = "0x000000000000C1CB11D5c062901F32D06248CE48"
        token_contract_metadata_mainnet = sourcify_client_mainnet.get_contract_metadata(
            partial_match_contract_address
        )
        self.assertEqual(token_contract_metadata_mainnet.name, "LiquidGasToken")
        self.assertIsInstance(token_contract_metadata_mainnet.abi, List)
        self.assertTrue(token_contract_metadata_mainnet.abi)
        self.assertTrue(token_contract_metadata_mainnet.partial_match)



from unittest import TestCase

from ...account_abstraction import UserOperationReceipt
from ..mocks.mock_bundler import user_operation_receipt_mock


class TestUserOperation(TestCase):
    def setUp(self):
        super().setUp()
        self.user_operation_receipt = UserOperationReceipt.from_bundler_response(
            user_operation_receipt_mock["result"]
        )

    def test_get_deployed_account(self):
        self.assertEqual(
            self.user_operation_receipt.get_deployed_account(),
            self.user_operation_receipt.sender,
        )

    def test_get_deposit(self):
        expected_value = 759_940_285_250_436
        self.assertEqual(self.user_operation_receipt.get_deposit(), expected_value)

    def test_get_module_address(self):
        expected_value = "0xa581c4A4DB7175302464fF3C06380BC3270b4037"
        self.assertEqual(
            self.user_operation_receipt.get_module_address(), expected_value
        )
